# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from spinn_utilities.progress_bar import ProgressBar
from spinnman.messages.scp.enums import Signal
from spinnman.model import ExecutableTargets
from spinnman.model.enums import CPUState
from spinn_front_end_common.utilities.helpful_functions import (
    flood_fill_binary_to_spinnaker)
from spinn_front_end_common.utilities.utility_objs import ExecutableType


class LoadExecutableImages(object):
    """ Go through the executable targets and load each binary to everywhere\
        and then send a start request to the cores that actually use it.
    """

    __slots__ = []

    def load_app_images(self, executable_targets, app_id, transceiver):
        """
        :param ~spinnman.model.ExecutableTargets executable_targets:
        :param int app_id:
        :param ~spinnman.transceiver.Transceiver transceiver:
        """
        self.__load_images(executable_targets, app_id, transceiver,
                           lambda ty: ty is not ExecutableType.SYSTEM,
                           "Loading executables onto the machine")

    def load_sys_images(self, executable_targets, app_id, transceiver):
        """
        :param ~spinnman.model.ExecutableTargets executable_targets:
        :param int app_id:
        :param ~spinnman.transceiver.Transceiver transceiver:
        """
        self.__load_images(executable_targets, app_id, transceiver,
                           lambda ty: ty is ExecutableType.SYSTEM,
                           "Loading system executables onto the machine")

    def __load_images(self, executable_targets, app_id, txrx, filt, label):
        """
        :param ~.ExecutableTargets executable_targets:
        :param int app_id:
        :param ~.Transceiver txrx:
        :param callable(ExecutableType,bool) filt:
        :param str label
        """
        # Compute what work is to be done here
        binaries, cores = self.filter_targets(executable_targets, filt)

        # ISSUE: Loading order may be non-constant on older Python
        progress = ProgressBar(cores.total_processors + 1, label)
        for binary in binaries:
            progress.update(flood_fill_binary_to_spinnaker(
                executable_targets, binary, txrx, app_id))

        self.__start_simulation(cores, txrx, app_id)
        progress.update()
        progress.end()

    @staticmethod
    def filter_targets(targets, filt):
        """
        :param ~spinnman.model.ExecutableTargets executable_targets:
        :param callable(ExecutableType,bool) filt:
        :rtype: tuple(list(str), ExecutableTargets)
        """
        binaries = []
        cores = ExecutableTargets()
        for exe_type in targets.executable_types_in_binary_set():
            if filt(exe_type):
                for aplx in targets.get_binaries_of_executable_type(exe_type):
                    binaries.append(aplx)
                    cores.add_subsets(
                        aplx, targets.get_cores_for_binary(aplx), exe_type)
        return binaries, cores

    @staticmethod
    def __start_simulation(executable_targets, txrx, app_id):
        """
        :param ~.ExecutableTargets executable_targets:
        :param ~.Transceiver txrx:
        :param int app_id:
        """
        txrx.wait_for_cores_to_be_in_state(
            executable_targets.all_core_subsets, app_id, [CPUState.READY])
        txrx.send_signal(app_id, Signal.START)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class Simulator_State(Enum):
    """ Different states the Simulator could be in.
    """
    #: init called
    INIT = (0, "init called")
    #: inside run method
    IN_RUN = (1, "inside run method")
    #: finished run method, but running forever
    RUN_FOREVER = (2, "finish run method but running forever")
    #: run ended shutdown not called
    FINISHED = (3, "run ended shutdown not called")
    #: shutdown called
    SHUTDOWN = (4, "shutdown called")
    #: stop requested in middle of run forever
    STOP_REQUESTED = (5, "stop requested in middle of run forever")

    def __new__(cls, value, doc=""):
        # pylint: disable=protected-access
        obj = object.__new__(cls)
        obj._value_ = value
        obj.__doc__ = doc
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



# Copyright (c) 2019-2020 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from pacman.model.graphs.machine import SourceSegmentedSDRAMMachinePartition
from spinn_front_end_common.utilities.exceptions import SpinnFrontEndException


class SDRAMOutgoingPartitionAllocator(object):

    def __call__(self, machine_graph, transceiver, placements, app_id):

        progress_bar = ProgressBar(
            total_number_of_things_to_do=len(machine_graph.vertices),
            string_describing_what_being_progressed=(
                "Allocating SDRAM for SDRAM outgoing egde partitions"))

        for machine_vertex in machine_graph.vertices:
            sdram_partitions = (
                machine_graph.get_sdram_edge_partitions_starting_at_vertex(
                    machine_vertex))
            for sdram_partition in sdram_partitions:

                # get placement, ones where the src is multiple,
                # you need to ask for the first pre vertex
                if isinstance(
                        sdram_partition, SourceSegmentedSDRAMMachinePartition):
                    placement = placements.get_placement_of_vertex(
                        next(iter(sdram_partition.pre_vertices)))
                else:
                    placement = placements.get_placement_of_vertex(
                        sdram_partition.pre_vertex)

                # total sdram
                total_sdram = (sdram_partition.total_sdram_requirements())

                # if bust, throw exception
                if total_sdram == 0:
                    raise SpinnFrontEndException(
                        "Cannot allocate sdram size of 0 for "
                        "partition {}".format(sdram_partition))

                # allocate
                sdram_base_address = transceiver.malloc_sdram(
                    placement.x, placement.y, total_sdram, app_id)

                # update
                sdram_partition.sdram_base_address = (sdram_base_address)

            progress_bar.update()
        progress_bar.end()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import numpy
import math
import scipy.stats
from spinn_utilities.log import FormatAdapter

logger = FormatAdapter(logging.getLogger(__name__))

# Define profiler time scale in ms
_MS_SCALE = (1.0 / 200000.0)

# Define the positions in tags
_START_TIME = 0
_DURATION = 1


class ProfileData(object):
    """ A container for profile data
    """
    # pylint: disable=assignment-from-no-return

    START_TIME = _START_TIME
    DURATION = _DURATION

    __slots__ = (
        # A dictionary of tag label to numpy array of start times and durations
        "_tags",

        # A list of tag labels indexed by the tag ID
        "_tag_labels",

        # The maximum time recorded
        "_max_time"
    )

    def __init__(self, tag_labels):
        """
        :param list(str) tag_labels: A list of labels indexed by tag ID
        """
        self._tag_labels = tag_labels
        self._tags = dict()
        self._max_time = None

    def add_data(self, data):
        """ Add profiling data read from the profile section

        :param bytearray data:
            Data read from the profile section on the machine
        """
        samples = numpy.asarray(data, dtype="uint8").view(dtype="<u4")

        # Slice data to separate times, tags and flags
        sample_times = samples[::2]
        sample_tags_and_flags = samples[1::2]

        # Further split the tags and flags word into separate arrays of tags
        # and flags
        sample_tags = numpy.bitwise_and(sample_tags_and_flags, 0x7FFFFFFF)
        sample_flags = numpy.right_shift(sample_tags_and_flags, 31)

        # Find indices of samples relating to entries and exits
        sample_entry_indices = numpy.where(sample_flags == 1)
        sample_exit_indices = numpy.where(sample_flags == 0)

        # Convert count-down times to count up times from 1st sample
        sample_times_ms = numpy.multiply(
            numpy.subtract(sample_times[0], sample_times),
            _MS_SCALE, dtype=numpy.float)

        # Slice tags and times into entry and exits
        entry_tags = sample_tags[sample_entry_indices]
        entry_times_ms = sample_times_ms[sample_entry_indices]
        exit_tags = sample_tags[sample_exit_indices]
        exit_times_ms = sample_times_ms[sample_exit_indices]

        # Loop through unique tags
        for tag in numpy.unique(sample_tags):
            self._add_tag_data(
                entry_tags, entry_times_ms, exit_tags, exit_times_ms, tag)

    def _add_tag_data(
            self, entry_tags, entry_times, exit_tags, exit_times, tag):
        """
        :param ~numpy.ndarray entry_tags:
        :param ~numpy.ndarray entry_times:
        :param ~numpy.ndarray exit_tags:
        :param ~numpy.ndarray exit_times:
        :param int tag:
        """
        # pylint: disable=too-many-arguments
        tag_label = self._tag_labels.get(tag, None)
        if tag_label is None:
            logger.warning("Unknown tag {} in profile data", tag)
            tag_label = "UNKNOWN"

        # Get indices where these tags occur
        tag_entry_indices = numpy.where(entry_tags == tag)
        tag_exit_indices = numpy.where(exit_tags == tag)

        # Use these to get subset for this tag
        tag_entry_times = entry_times[tag_entry_indices]
        tag_exit_times = exit_times[tag_exit_indices]

        # If the first exit is before the first
        # Entry, add a dummy entry at beginning
        if tag_exit_times[0] < tag_entry_times[0]:
            logger.warning("Profile starts mid-tag")
            tag_entry_times = numpy.append(0.0, tag_entry_times)

        if len(tag_entry_times) > len(tag_exit_times):
            logger.warning("profile finishes mid-tag")
            tag_entry_times = tag_entry_times[
                :len(tag_exit_times) - len(tag_entry_times)]

        # Subtract entry times from exit times to get durations of each
        # call in ms
        tag_durations = numpy.subtract(tag_exit_times, tag_entry_times)

        # Add entry times and durations to dictionary
        self._tags[tag_label] = (tag_entry_times, tag_durations)

        # Keep track of the maximum time
        self._max_time = numpy.max(tag_entry_times + tag_durations)

    @property
    def tags(self):
        """ The tags recorded as labels

        :rtype: list(str)
        """
        return self._tags.keys()

    def get_mean_ms(self, tag):
        """ Get the mean time in milliseconds spent on operations with the
            given tag

        :param str tag: The tag to get the mean time for
        :rtype: float
        """
        return numpy.average(self._tags[tag][_DURATION])

    def get_n_calls(self, tag):
        """ Get the number of times the given tag was recorded

        :param str tag: The tag to get the number of calls of
        :rtype: int
        """
        return self._tags[tag][_DURATION].size

    def get_mean_n_calls_per_ts(self, tag, machine_time_step_ms):
        """ Get the mean number of times the given tag was recorded per\
            timestep

        :param str tag: The tag to get the data for
        :param float machine_time_step_ms:
            The time step of the simulation in microseconds
        :rtype: float
        """
        n_points = math.ceil(
            self._max_time / machine_time_step_ms)
        endpoint = n_points * machine_time_step_ms
        bins = numpy.linspace(0, endpoint, n_points + 1)
        return numpy.average(numpy.histogram(
            self._tags[tag][_START_TIME], bins)[0])

    def get_mean_ms_per_ts(self, tag, machine_time_step_ms):
        """ Get the mean time in milliseconds spent on operations with the\
            given tag per timestep

        :param str tag: The tag to get the data for
        :param float machine_time_step_ms:
            The time step of the simulation in microseconds
        :rtype: float
        """
        n_points = math.ceil(
            self._max_time / machine_time_step_ms)
        endpoint = n_points * machine_time_step_ms
        bins = numpy.linspace(0, endpoint, n_points + 1)
        mean_per_ts = scipy.stats.binned_statistic(
            self._tags[tag][_START_TIME], self._tags[tag][_DURATION],
            "mean", bins).statistic
        mean_per_ts[numpy.isnan(mean_per_ts)] = 0
        return numpy.average(
            mean_per_ts[numpy.logical_not(numpy.isnan(mean_per_ts))])



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.utilities.constants import (
    MICRO_TO_MILLISECOND_CONVERSION)
from spinn_front_end_common.interface.profiling import AbstractHasProfileData


class ProfileDataGatherer(object):
    """ Gets all the profiling data recorded by vertices and writes it to\
        files.
    """

    __slots__ = []

    def __call__(
            self, transceiver, placements, provenance_file_path,
            machine_time_step):
        """
        :param ~spinnman.transceiver.Transceiver transceiver:
            the SpiNNMan interface object
        :param ~pacman.model.placements.Placements placements:
            The placements of the vertices
        :param str provenance_file_path:
            The location to store the profile data
        :param int machine_time_step:
            machine time step in ms
        """
        # pylint: disable=too-many-arguments
        machine_time_step_ms = (
            float(machine_time_step) / MICRO_TO_MILLISECOND_CONVERSION)

        progress = ProgressBar(
            placements.n_placements, "Getting profile data")

        # retrieve provenance data from any cores that provide data
        for placement in progress.over(placements.placements):
            if isinstance(placement.vertex, AbstractHasProfileData):
                # get data
                profile_data = placement.vertex.get_profile_data(
                    transceiver, placement)
                if profile_data.tags:
                    self._write(placement, profile_data, machine_time_step_ms,
                                provenance_file_path)

    _FMT_A = "{: <{}s} {: <7s} {: <14s} {: <14s} {: <14s}\n"
    _FMT_B = "{:-<{}s} {:-<7s} {:-<14s} {:-<14s} {:-<14s}\n"
    _FMT_C = "{: <{}s} {: >7d} {: >14.6f} {: >14.6f} {: >14.6f}\n"

    @classmethod
    def _write(cls, p, profile_data, machine_time_step_ms, directory):
        """
        :param ~.Placement p:
        :param ProfileData profile_data:
        :param float machine_time_step_ms:
        :param str directory:
        """
        max_tag_len = max(len(tag) for tag in profile_data.tags)
        file_name = os.path.join(
            directory, "{}_{}_{}_profile.txt".format(p.x, p.y, p.p))

        # write profile data to file, creating if necessary
        with open(file_name, "a") as f:
            # Write header
            f.write(cls._FMT_A.format(
                "tag", max_tag_len, "n_calls", "mean_ms", "n_calls_per_ts",
                "mean_ms_per_ts"))
            f.write(cls._FMT_B.format("", max_tag_len, "", "", "", ""))

            # Write content
            for tag in profile_data.tags:
                f.write(cls._FMT_C.format(
                    tag, max_tag_len, profile_data.get_n_calls(tag),
                    profile_data.get_mean_ms(tag),
                    profile_data.get_mean_n_calls_per_ts(
                        tag, machine_time_step_ms),
                    profile_data.get_mean_ms_per_ts(
                        tag, machine_time_step_ms)))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from pacman.model.constraints.placer_constraints import ChipAndCoreConstraint
from spinn_front_end_common.utility_models import (
    ChipPowerMonitor, ChipPowerMonitorMachineVertex)

_LABEL = "chip_power_monitor_{}_vertex_for_chip({}:{})"


class InsertChipPowerMonitorsToGraphs(object):
    """ Adds chip power monitors into a given graph.
    """

    def __call__(
            self, machine, machine_graph, n_samples_per_recording,
            sampling_frequency, application_graph=None):
        """ Adds chip power monitor vertices on Ethernet connected chips as\
            required.

        :param ~spinn_machine.Machine machine:
            the SpiNNaker machine as discovered
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            the machine graph
        :param int n_samples_per_recording:
        :param int sampling_frequency:
        :param application_graph: the application graph
        :type application_graph:
            ~pacman.model.graphs.application.ApplicationGraph
        """
        # pylint: disable=too-many-arguments

        # create progress bar
        progress = ProgressBar(
            machine.n_chips, "Adding Chip power monitors to Graph")

        if application_graph is not None:
            self.__add_app(
                application_graph, machine_graph, machine,
                n_samples_per_recording, sampling_frequency, progress)
        else:
            self.__add_mach_only(
                machine_graph, machine,
                n_samples_per_recording, sampling_frequency, progress)

    @staticmethod
    def __add_app(
            application_graph, machine_graph, machine, n_samples_per_recording,
            sampling_frequency, progress):
        app_vertex = ChipPowerMonitor(
            label="ChipPowerMonitor",
            sampling_frequency=sampling_frequency,
            n_samples_per_recording=n_samples_per_recording)
        application_graph.add_vertex(app_vertex)
        for chip in progress.over(machine.chips):
            if not chip.virtual:
                machine_graph.add_vertex(app_vertex.create_machine_vertex(
                    vertex_slice=None, resources_required=None,
                    label=_LABEL.format("machine", chip.x, chip.y),
                    constraints=[ChipAndCoreConstraint(chip.x, chip.y)]))

    @staticmethod
    def __add_mach_only(
            machine_graph, machine, n_samples_per_recording,
            sampling_frequency, progress):
        for chip in progress.over(machine.chips):
            if not chip.virtual:
                machine_graph.add_vertex(ChipPowerMonitorMachineVertex(
                    label=_LABEL.format("machine", chip.x, chip.y),
                    constraints=[ChipAndCoreConstraint(chip.x, chip.y)],
                    app_vertex=None,
                    sampling_frequency=sampling_frequency,
                    n_samples_per_recording=n_samples_per_recording))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


class AbstractDatabase(object, metaclass=AbstractBase):
    """
    This API separates the required database calls from the implementation.

    Methods here are designed for the convenience of the caller not the\
    database.

    There should only ever be a single Database Object in use at any time.\
    In the case of application_graph_changed the first should closed and\
    a new one created.

    Do not assume that just because 2 database objects where opened with the\
    same parameters (for example SQLite file)\
    that they hold the same data. \
    In fact the second init is allowed to delete any previous data.

    While not recommended implementation objects are allowed to hold data in \
    memory, with the exception of data required by the java which must be \
    in the database once commit is called.
    """

    __slots__ = ()

    @abstractmethod
    def close(self):
        """ Signals that the database can be closed and will not be reused.

        Once this is called any other method in this API is allowed to\
        raise any kind of exception.
        """

    @abstractmethod
    def store_data_in_region_buffer(self, x, y, p, region, missing, data):
        """ Store some information in the corresponding buffer for a\
            specific chip, core and recording region.

        :param int x: x coordinate of the chip
        :param int y: y coordinate of the chip
        :param int p: Core within the specified chip
        :param int region: Region containing the data to be stored
        :param bool missing: Whether any data is missing
        :param bytearray data: data to be stored

            .. note::
                Implementations may assume this to be shorter than 1GB
        """

    @abstractmethod
    def get_region_data(self, x, y, p, region):
        """ Get the data stored for a given region of a given core

        :param int x: x coordinate of the chip
        :param int y: y coordinate of the chip
        :param int p: Core within the specified chip
        :param int region: Region containing the data
        :return: a buffer containing all the data received during the\
            simulation, and a flag indicating if any data was missing

            .. note::
                Implementations should not assume that the total buffer is
                necessarily shorter than 1GB.

        :rtype: tuple(memoryview, bool)
        """

    @abstractmethod
    def clear(self):
        """ Clears the data for all regions.

        .. note::
            This method will be removed when the database moves to
            keeping data after reset.

        :rtype: None
        """

    @abstractmethod
    def clear_region(self, x, y, p, region):
        """ Clears the data for a single region.

        .. note::
            This method *loses information!*

        :param int x: x coordinate of the chip
        :param int y: y coordinate of the chip
        :param int p: Core within the specified chip
        :param int region: Region containing the data to be cleared
        :return: True if any region was changed
        :rtype: bool
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
import logging
from enum import IntEnum
import numpy
from data_specification.enums import DataType
from pacman.executor.injection_decorator import (
    inject_items, supports_injection)
from pacman.model.graphs.machine import MachineVertex
from pacman.model.resources import (
    CPUCyclesPerTickResource, DTCMResource, ResourceContainer, VariableSDRAM)
from spinn_front_end_common.abstract_models import (
    AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary)
from spinn_front_end_common.interface.buffer_management import (
    recording_utilities)
from spinn_front_end_common.interface.buffer_management.buffer_models import (
    AbstractReceiveBuffersToHost)
from spinn_front_end_common.utilities import globals_variables
from spinn_front_end_common.utilities.constants import (
    SYSTEM_BYTES_REQUIREMENT, SIMULATION_N_BYTES, BYTES_PER_WORD)
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement)
from spinn_front_end_common.interface.simulation.simulation_utilities import (
    get_simulation_header_array)

logger = FormatAdapter(logging.getLogger(__name__))
BINARY_FILE_NAME = "chip_power_monitor.aplx"

RECORDING_SIZE_PER_ENTRY = 18 * BYTES_PER_WORD
DEFAULT_MALLOCS_USED = 3
CONFIG_SIZE_IN_BYTES = 2 * BYTES_PER_WORD


@supports_injection
class ChipPowerMonitorMachineVertex(
        MachineVertex, AbstractHasAssociatedBinary,
        AbstractGeneratesDataSpecification, AbstractReceiveBuffersToHost):
    """ Machine vertex for C code representing functionality to record\
        idle times in a machine graph.
    """
    __slots__ = ["_n_samples_per_recording", "_sampling_frequency"]

    class _REGIONS(IntEnum):
        # data regions
        SYSTEM = 0
        CONFIG = 1
        RECORDING = 2

    #: which channel in the recording region has the recorded samples
    _SAMPLE_RECORDING_CHANNEL = 0

    def __init__(
            self, label, constraints, n_samples_per_recording,
            sampling_frequency, app_vertex=None, vertex_slice=None):
        """
        :param str label: vertex label
        :param constraints: constraints on this vertex
        :type constraints:
            iterable(~pacman.model.constraints.AbstractConstraint)
        :param int n_samples_per_recording:
            how may samples between recording entry
        :param int sampling_frequency: how often to sample, in microseconds
        :param ChipPowerMonitor app_vertex: associated application vertex
        :param ~pacman.model.graphs.common.Slice vertex_slice:
        """
        super().__init__(
            label=label, constraints=constraints, app_vertex=app_vertex,
            vertex_slice=vertex_slice)
        self._n_samples_per_recording = n_samples_per_recording
        self._sampling_frequency = sampling_frequency

    @property
    def sampling_frequency(self):
        """ how often to sample, in microseconds

        :rtype: int
        """
        return self._sampling_frequency

    @property
    def n_samples_per_recording(self):
        """ how may samples to take between making recording entries

        :rtype: int
        """
        return self._n_samples_per_recording

    @property
    @overrides(MachineVertex.resources_required)
    def resources_required(self):
        # pylint: disable=arguments-differ
        sim = globals_variables.get_simulator()
        return self.get_resources(
            sim.machine_time_step, sim.time_scale_factor,
            self._n_samples_per_recording, self._sampling_frequency)

    @staticmethod
    def get_resources(
            time_step, time_scale_factor,
            n_samples_per_recording, sampling_frequency):
        """ Get the resources used by this vertex

        :param int time_step:
        :param int time_scale_factor:
        :param int n_samples_per_recording:
        :param float sampling_frequency:
        :rtype: ~pacman.model.resources.ResourceContainer
        """
        # pylint: disable=too-many-locals
        step_in_microseconds = (time_step * time_scale_factor)
        # The number of sample per step CB believes does not have to be an int
        samples_per_step = (step_in_microseconds / sampling_frequency)
        recording_per_step = (samples_per_step / n_samples_per_recording)
        max_recording_per_step = math.ceil(recording_per_step)
        overflow_recordings = max_recording_per_step - recording_per_step
        system = SYSTEM_BYTES_REQUIREMENT
        config = CONFIG_SIZE_IN_BYTES
        recording = recording_utilities.get_recording_header_size(1)
        recording += recording_utilities.get_recording_data_constant_size(1)
        fixed_sdram = system + config + recording
        with_overflow = (
            fixed_sdram + overflow_recordings * RECORDING_SIZE_PER_ENTRY)
        per_timestep = recording_per_step * RECORDING_SIZE_PER_ENTRY

        return ResourceContainer(
            sdram=VariableSDRAM(with_overflow, per_timestep),
            cpu_cycles=CPUCyclesPerTickResource(100),
            dtcm=DTCMResource(100))

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return BINARY_FILE_NAME

    @staticmethod
    def binary_file_name():
        """ Return the string binary file name

        :rtype: str
        """
        return BINARY_FILE_NAME

    @inject_items({"time_scale_factor": "TimeScaleFactor",
                   "machine_time_step": "MachineTimeStep",
                   "data_n_time_steps": "DataNTimeSteps"})
    @overrides(AbstractGeneratesDataSpecification.generate_data_specification,
               additional_arguments={
                   "machine_time_step", "time_scale_factor",
                   "data_n_time_steps"})
    def generate_data_specification(
            self, spec, placement,  # @UnusedVariable
            machine_time_step, time_scale_factor, data_n_time_steps):
        """ Supports the application vertex calling this directly

        :param ~data_specification.DataSpecificationGenerator spec: data spec
        :param int machine_time_step: machine time step
        :param int time_scale_factor: time scale factor
        :param int data_n_time_steps: timesteps to reserve data for
        """
        # pylint: disable=too-many-arguments
        spec.comment("\n*** Spec for ChipPowerMonitor Instance ***\n\n")

        # Construct the data images needed for the Neuron:
        self._reserve_memory_regions(spec)
        self._write_setup_info(
            spec, machine_time_step, time_scale_factor, data_n_time_steps)
        self._write_configuration_region(spec)

        # End-of-Spec:
        spec.end_specification()

    def _write_configuration_region(self, spec):
        """ Write the data needed by the C code to configure itself

        :param ~data_specification.DataSpecificationGenerator spec:
            spec object
        """
        spec.switch_write_focus(region=self._REGIONS.CONFIG)
        spec.write_value(self._n_samples_per_recording,
                         data_type=DataType.UINT32)
        spec.write_value(self._sampling_frequency, data_type=DataType.UINT32)

    def _write_setup_info(
            self, spec, machine_time_step, time_scale_factor,
            n_machine_time_steps):
        """ Writes the system data as required.

        :param ~data_specification.DataSpecificationGenerator spec:
            the DSG spec writer
        :param int machine_time_step: the machine time step
        :param int time_scale_factor: the time scale factor
        """
        # pylint: disable=too-many-arguments
        spec.switch_write_focus(region=self._REGIONS.SYSTEM)
        spec.write_array(get_simulation_header_array(
            self.get_binary_file_name(), machine_time_step, time_scale_factor))

        spec.switch_write_focus(region=self._REGIONS.RECORDING)
        recorded_region_sizes = [
            self._deduce_sdram_requirements_per_timer_tick(
                machine_time_step, time_scale_factor) * n_machine_time_steps]
        spec.write_array(recording_utilities.get_recording_header_array(
            recorded_region_sizes))

    def _reserve_memory_regions(self, spec):
        """ Reserve the DSG memory regions as required

        :param ~data_specification.DataSpecificationGenerator spec:
            the DSG specification to reserve in
        """
        spec.comment("\nReserving memory space for data regions:\n\n")

        # Reserve memory:
        spec.reserve_memory_region(
            region=self._REGIONS.SYSTEM,
            size=SIMULATION_N_BYTES,
            label='system')
        spec.reserve_memory_region(
            region=self._REGIONS.CONFIG,
            size=CONFIG_SIZE_IN_BYTES, label='config')
        spec.reserve_memory_region(
            region=self._REGIONS.RECORDING,
            size=recording_utilities.get_recording_header_size(1),
            label="Recording")

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return self.binary_start_type()

    @staticmethod
    def binary_start_type():
        """ The type of binary that implements this vertex

        :return: starttype enum
        :rtype: ExecutableType
        """
        return ExecutableType.USES_SIMULATION_INTERFACE

    @overrides(AbstractReceiveBuffersToHost.get_recording_region_base_address)
    def get_recording_region_base_address(self, txrx, placement):
        return locate_memory_region_for_placement(
            placement, self._REGIONS.RECORDING, txrx)

    @overrides(AbstractReceiveBuffersToHost.get_recorded_region_ids)
    def get_recorded_region_ids(self):
        return [0]

    def _deduce_sdram_requirements_per_timer_tick(
            self, machine_time_step, time_scale_factor):
        """ Deduce SDRAM usage per timer tick

        :param int machine_time_step: the machine time step
        :param float time_scale_factor: the time scale factor
        :return: the SDRAM usage
        :rtype: int
        """
        timer_tick_in_micro_seconds = machine_time_step * time_scale_factor
        recording_time = \
            self._sampling_frequency * self._n_samples_per_recording
        n_entries = math.floor(timer_tick_in_micro_seconds / recording_time)
        return int(math.ceil(n_entries * RECORDING_SIZE_PER_ENTRY))

    def get_recorded_data(self, placement, buffer_manager):
        """ Get data from SDRAM given placement and buffer manager

        :param ~pacman.model.placements.Placement placement:
            the location on machine to get data from
        :param BufferManager buffer_manager:
            the buffer manager that might have data
        :return: results, an array with 1 dimension of uint32 values
        :rtype: ~numpy.ndarray
        """
        # for buffering output info is taken form the buffer manager
        # get raw data as a byte array
        record_raw, data_missing = buffer_manager.get_data_by_placement(
            placement, self._SAMPLE_RECORDING_CHANNEL)
        if data_missing:
            logger.warning(
                "Chip Power monitor has lost data on chip({}, {})",
                placement.x, placement.y)

        results = (
            numpy.frombuffer(record_raw, dtype="uint32").reshape(-1, 18) /
            self.n_samples_per_recording)
        return results



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import abstractproperty
from spinn_utilities.overrides import overrides
from .abstract_sends_buffers_from_host import AbstractSendsBuffersFromHost


class SendsBuffersFromHostPreBufferedImpl(
        AbstractSendsBuffersFromHost, allow_derivation=True):
    """ Implementation of :py:class:`AbstractReceiveBuffersToHost`\
        which uses an existing set of buffers for the details.
    """
    # pylint: disable=unsubscriptable-object, no-member

    __slots__ = ()

    @abstractproperty
    def send_buffers(self):
        """
        :rtype: dict(int,
            ~spinn_front_end_common.interface.buffer_management.storage_objects.BufferedSendingRegion)
        """

    @overrides(AbstractSendsBuffersFromHost.buffering_input)
    def buffering_input(self):
        return self.send_buffers is not None

    @overrides(AbstractSendsBuffersFromHost.get_regions)
    def get_regions(self):
        return self.send_buffers.keys()

    @overrides(AbstractSendsBuffersFromHost.is_next_timestamp)
    def is_next_timestamp(self, region):
        return self.send_buffers[region].is_next_timestamp

    @overrides(AbstractSendsBuffersFromHost.get_next_timestamp)
    def get_next_timestamp(self, region):
        return self.send_buffers[region].next_timestamp

    @overrides(AbstractSendsBuffersFromHost.is_next_key)
    def is_next_key(self, region, timestamp):
        return self.send_buffers[region].is_next_key(timestamp)

    @overrides(AbstractSendsBuffersFromHost.get_next_key)
    def get_next_key(self, region):
        return self.send_buffers[region].next_key

    @overrides(AbstractSendsBuffersFromHost.is_empty)
    def is_empty(self, region):
        return len(self.send_buffers[region].timestamps) == 0

    @overrides(AbstractSendsBuffersFromHost.rewind)
    def rewind(self, region):
        self.send_buffers[region].rewind()



# Copyright (c) 2019-2020 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import logging
import struct
from collections import defaultdict
from spinn_utilities.abstract_base import AbstractBase, abstractproperty
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_utilities.progress_bar import ProgressBar
from spinn_machine import CoreSubsets, Router
from spinnman.exceptions import (
    SpinnmanInvalidParameterException,
    SpinnmanUnexpectedResponseCodeException, SpiNNManCoresNotInStateException)
from spinnman.model import ExecutableTargets
from spinnman.model.enums import CPUState
from pacman.model.routing_tables import MulticastRoutingTables
from pacman.operations.router_compressors.ordered_covering_router_compressor\
    import (
        get_generality as
        ordered_covering_generality)
from spinn_front_end_common.abstract_models.\
    abstract_supports_bit_field_routing_compression import (
        AbstractSupportsBitFieldRoutingCompression)
from spinn_front_end_common.utilities.report_functions.\
    bit_field_compressor_report import (
        generate_provenance_item)
from spinn_front_end_common.utilities.exceptions import (
    CantFindSDRAMToUseException)
from spinn_front_end_common.utilities.helpful_functions import (
    get_defaultable_source_id, n_word_struct)
from spinn_front_end_common.utilities.system_control_logic import (
    run_system_application)
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from .load_executable_images import LoadExecutableImages
from .host_bit_field_router_compressor import HostBasedBitFieldRouterCompressor

logger = FormatAdapter(logging.getLogger(__name__))

#: sdram allocation for addresses
SIZE_OF_SDRAM_ADDRESS_IN_BYTES = (17 * 2 * 4) + (3 * 4)

# 7 pointers or int for each core. 4 Bytes for each  18 cores max
SIZE_OF_COMMS_SDRAM = 7 * 4 * 18

SECOND_TO_MICRO_SECOND = 1000000


class MachineBitFieldRouterCompressor(object, metaclass=AbstractBase):
    """ On-machine bitfield-aware routing table compression.
    """

    __slots__ = []

    #: sdram tag the router compressor expects to find there routing tables in
    ROUTING_TABLE_SDRAM_TAG = 1

    #: sdram tag for the addresses the router compressor expects to find the
    #: bitfield addresses for the chip.
    BIT_FIELD_ADDRESSES_SDRAM_TAG = 2

    #
    TIMES_CYCLED_ROUTING_TABLES = 3

    #: the successful identifier
    SUCCESS = 0

    #: how many header elements are in the region addresses (1, n addresses)
    N_REGIONS_ELEMENT = 1

    # the number of bytes needed to read the user 2 register
    _USER_BYTES = 4

    #: min size a heap object needs in sdram. (limit on the size of useful
    #: sdram regions to steal)
    _MIN_SIZE_FOR_HEAP = 32

    # bit offset for compress only when needed
    _ONLY_WHEN_NEEDED_BIT_OFFSET = 1

    # bit offset for compress as much as possible
    _AS_MUCH_AS_POSS_BIT_OFFSET = 2

    # structs for performance requirements.
    _FOUR_WORDS = struct.Struct("<IIII")
    _TWO_WORDS = struct.Struct("<II")
    _ONE_WORD = struct.Struct("<I")

    # binary names
    _BIT_FIELD_SORTER_AND_SEARCH_EXECUTOR_APLX = \
        "bit_field_sorter_and_searcher.aplx"

    _PROGRESS_BAR_TEXT = \
        "on chip {} compressor with bitfields"

    _HOST_BAR_TEXT = \
        "on host compressing routing tables and merging in bitfields as " \
        "appropriate"
    _ON_CHIP_ERROR_MESSAGE = \
        "The router compressor with bit field on {}:{} failed to complete. " \
        "Will execute host based routing compression instead"

    _ON_HOST_WARNING_MESSAGE = \
        "Will be executing compression for {} chips on the host, as they " \
        "failed to complete when running on chip"

    def __call__(
            self, routing_tables, transceiver, machine, app_id,
            provenance_file_path, machine_graph, placements, executable_finder,
            write_compressor_iobuf, produce_report, default_report_folder,
            target_length, routing_infos, time_to_try_for_each_iteration,
            use_timer_cut_off, machine_time_step, time_scale_factor,
            threshold_percentage, retry_count, executable_targets,
            compress_as_much_as_possible=False, provenance_data_objects=None):
        """ entrance for routing table compression with bit field

        :param routing_tables: routing tables
        :type routing_tables:
            ~pacman.model.routing_tables.MulticastRoutingTables
        :param ~spinnman.transceiver.Transceiver transceiver: spinnman instance
        :param ~spinn_machine.Machine machine: spinnMachine instance
        :param int app_id: app id of the application
        :param str provenance_file_path: file path for prov data
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            machine graph
        :param ~pacman.model.placements.Placements placements:
            placements on machine
        :param ExecutableFinder executable_finder:
            where are binaries are located
        :param bool write_compressor_iobuf: flag saying if read IOBUF
        :param bool produce_report:
        :param str default_report_folder:
        :param int target_length:
        :param ~pacman.model.routing_info.RoutingInfo routing_infos:
        :param int time_to_try_for_each_iteration:
        :param bool use_timer_cut_off:
        :param int machine_time_step:
        :param int time_scale_factor:
        :param int threshold_percentage:
            the percentage of bitfields to do on chip
            before its considered a success
        :param int retry_count:
            Number of times that the sorters should set of the compressions
            again
        :param ~spinnman.model.ExecutableTargets executable_targets:
            the set of targets and executables
        :param bool compress_as_much_as_possible:
            whether to compress as much as possible
        :param list(ProvenanceDataItem) provenance_data_objects:
        :return: where the compressors ran, and the provenance they generated
        :rtype: tuple(~spinnman.model.ExecutableTargets,
            list(ProvenanceDataItem))
        """

        # build provenance data objects
        if provenance_data_objects is not None:
            prov_items = provenance_data_objects
        else:
            prov_items = list()

        if len(routing_tables.routing_tables) == 0:
            return ExecutableTargets(), prov_items

        # new app id for this simulation
        routing_table_compressor_app_id = \
            transceiver.app_id_tracker.get_new_id()

        text = self._PROGRESS_BAR_TEXT.format(self.compressor_type)
        if retry_count is not None:
            text += " capped at {} retries".format(retry_count)
        progress_bar = ProgressBar(
            total_number_of_things_to_do=(
                len(machine_graph.vertices) +
                (len(routing_tables.routing_tables) *
                 self.TIMES_CYCLED_ROUTING_TABLES)),
            string_describing_what_being_progressed=text)

        # locate data and on_chip_cores to load binary on
        (addresses, matrix_addresses_and_size) = self._generate_addresses(
            machine_graph, placements, transceiver, progress_bar)

        # create executable targets
        (compressor_executable_targets, bit_field_sorter_executable_path,
         bit_field_compressor_executable_path) = self._generate_core_subsets(
            routing_tables, executable_finder, machine, progress_bar,
            executable_targets)

        # load data into sdram
        on_host_chips = self._load_data(
            addresses, transceiver, routing_table_compressor_app_id,
            routing_tables, app_id, machine,
            compress_as_much_as_possible, progress_bar,
            compressor_executable_targets,
            matrix_addresses_and_size, time_to_try_for_each_iteration,
            bit_field_compressor_executable_path,
            bit_field_sorter_executable_path, threshold_percentage,
            retry_count)

        # load and run binaries
        try:
            run_system_application(
                compressor_executable_targets,
                routing_table_compressor_app_id, transceiver,
                provenance_file_path, executable_finder,
                write_compressor_iobuf,
                functools.partial(
                    self._check_bit_field_router_compressor_for_success,
                    host_chips=on_host_chips,
                    sorter_binary_path=bit_field_sorter_executable_path,
                    prov_data_items=prov_items),
                [CPUState.FINISHED], True,
                "bit_field_compressor_on_{}_{}_{}.txt",
                [bit_field_sorter_executable_path], progress_bar,
                logger=logger)
        except SpiNNManCoresNotInStateException as e:
            logger.exception(transceiver.get_core_status_string(
                e.failed_core_states()))
            raise e

        # start the host side compressions if needed
        if len(on_host_chips) != 0:
            logger.warning(self._ON_HOST_WARNING_MESSAGE, len(on_host_chips))
            progress_bar = ProgressBar(
                total_number_of_things_to_do=len(on_host_chips),
                string_describing_what_being_progressed=self._HOST_BAR_TEXT)
            host_compressor = HostBasedBitFieldRouterCompressor()
            compressed_pacman_router_tables = MulticastRoutingTables()

            key_atom_map = host_compressor.generate_key_to_atom_map(
                machine_graph, routing_infos)

            for (chip_x, chip_y) in progress_bar.over(on_host_chips, False):
                report_folder_path = host_compressor.generate_report_path(
                    default_report_folder)
                prov_items.append(
                    host_compressor.start_compression_selection_process(
                        router_table=routing_tables.get_routing_table_for_chip(
                            chip_x, chip_y),
                        produce_report=produce_report,
                        report_folder_path=report_folder_path,
                        transceiver=transceiver, machine_graph=machine_graph,
                        placements=placements, machine=machine,
                        target_length=target_length,
                        compressed_pacman_router_tables=(
                            compressed_pacman_router_tables),
                        key_atom_map=key_atom_map))

            # load host compressed routing tables
            for table in compressed_pacman_router_tables.routing_tables:
                if (not machine.get_chip_at(table.x, table.y).virtual
                        and table.multicast_routing_entries):
                    transceiver.clear_multicast_routes(table.x, table.y)
                    transceiver.load_multicast_routes(
                        table.x, table.y, table.multicast_routing_entries,
                        app_id=app_id)

            progress_bar.end()

        return compressor_executable_targets, prov_items

    @abstractproperty
    def compressor_aplx(self):
        """

        :return: The name of the compressor aplx file to use
        """

    @abstractproperty
    def compressor_type(self):
        """

        :return: The name of the compressor (excluding bitfields) being used
        """
    def _generate_core_subsets(
            self, routing_tables, executable_finder, machine, progress_bar,
            system_executable_targets):
        """ generates the core subsets for the binaries

        :param ~.MulticastRoutingTables routing_tables: the routing tables
        :param ~.ExecutableFinder executable_finder: the executable path finder
        :param ~.Machine machine: the spinn machine instance
        :param ~.ProgressBar progress_bar: progress bar
        :param ExecutableTargets system_executable_targets:
            the executables targets to cores
        :return: (targets, sorter path, and compressor path)
        :rtype: tuple(ExecutableTargets, str, str)
        """
        bit_field_sorter_cores = CoreSubsets()
        bit_field_compressor_cores = CoreSubsets()

        _, cores = LoadExecutableImages.filter_targets(
            system_executable_targets, lambda ty: ty is ExecutableType.SYSTEM)

        for routing_table in progress_bar.over(routing_tables, False):
            # add 1 core to the sorter, and the rest to compressors
            sorter = None
            for processor in machine.get_chip_at(
                    routing_table.x, routing_table.y).processors:
                if (not processor.is_monitor and
                        not cores.all_core_subsets.is_core(
                            routing_table.x, routing_table.y,
                            processor.processor_id)):
                    if sorter is None:
                        sorter = processor
                        bit_field_sorter_cores.add_processor(
                            routing_table.x, routing_table.y,
                            processor.processor_id)
                    else:
                        bit_field_compressor_cores.add_processor(
                            routing_table.x, routing_table.y,
                            processor.processor_id)

        # convert core subsets into executable targets
        executable_targets = ExecutableTargets()

        # bit field executable paths
        bit_field_sorter_executable_path = \
            executable_finder.get_executable_path(
                self._BIT_FIELD_SORTER_AND_SEARCH_EXECUTOR_APLX)

        bit_field_compressor_executable_path = \
            executable_finder.get_executable_path(self.compressor_aplx)

        # add the sets
        executable_targets.add_subsets(
            binary=bit_field_sorter_executable_path,
            subsets=bit_field_sorter_cores,
            executable_type=ExecutableType.SYSTEM)
        executable_targets.add_subsets(
            binary=bit_field_compressor_executable_path,
            subsets=bit_field_compressor_cores,
            executable_type=ExecutableType.SYSTEM)

        return (executable_targets, bit_field_sorter_executable_path,
                bit_field_compressor_executable_path)

    def _check_bit_field_router_compressor_for_success(
            self, executable_targets, transceiver, host_chips,
            sorter_binary_path, prov_data_items):
        """ Goes through the cores checking for cores that have failed to\
            generate the compressed routing tables with bitfield

        :param ExecutableTargets executable_targets:
            cores to load router compressor with bitfield on
        :param ~.Transceiver transceiver: SpiNNMan instance
        :param list(tuple(int,int)) host_chips:
            the chips which need to be ran on host.
        :param str sorter_binary_path: the path to the sorter binary
        :param list(ProvenanceDataItem) prov_data_items:
            the store of data items
        :rtype: bool
        """
        sorter_cores = executable_targets.get_cores_for_binary(
            sorter_binary_path)
        for core_subset in sorter_cores:
            x = core_subset.x
            y = core_subset.y

            for p in core_subset.processor_ids:

                # Read the result from USER1/USER2 registers
                user_1_base_address = \
                    transceiver.get_user_1_register_address_from_core(p)
                user_2_base_address = \
                    transceiver.get_user_2_register_address_from_core(p)
                result = transceiver.read_word(x, y, user_1_base_address)
                bit_fields_merged = transceiver.read_word(
                    x, y, user_2_base_address)

                if result != self.SUCCESS:
                    if (x, y) not in host_chips:
                        host_chips.append((x, y))
                    return False
                prov_data_items.append(generate_provenance_item(
                    x, y, bit_fields_merged))
        return True

    def _load_data(
            self, addresses, transceiver, routing_table_compressor_app_id,
            routing_tables, app_id, machine,
            compress_as_much_as_possible, progress_bar, cores,
            matrix_addresses_and_size, time_per_iteration,
            bit_field_compressor_executable_path,
            bit_field_sorter_executable_path, threshold_percentage,
            retry_count):
        """ load all data onto the chip

        :param dict(tuple(int,int),tuple(int,int)) addresses:
            the addresses for bitfields in sdram
        :param ~.Transceiver transceiver: the spinnMan instance
        :param routing_table_compressor_app_id: the app id for the system app
        :param ~.MulticastRoutingTables routing_tables:
            the routing tables
        :param int app_id: the appid of the application
        :param ~.ProgressBar progress_bar: progress bar
        :param bool compress_as_much_as_possible:
            whether to compress as much as possible
        :param ExecutableTargets cores:
            the cores that compressor will run on
        :param dict matrix_addresses_and_size:
            maps chips to regeneration sdram and size for exploitation
        :param str bit_field_compressor_executable_path:
            the path to the compressor binary path
        :param str bit_field_sorter_executable_path:
            the path to the sorter binary
        :param int threshold_percentage:
            the percentage of bitfields the user has defined as a minimum
            needed to pass to be successful.
        :param retry_count:
            Number of times that the sorters should set of the compressions
            again. None for as much as needed
        :type retry_count: int or None
        :return:
            the list of tuples saying which chips this will need to use
            host compression, as the malloc failed.
        :rtype: list(tuple(int,int))
        """
        run_by_host = list()
        for table in routing_tables.routing_tables:
            if not machine.get_chip_at(table.x, table.y).virtual:
                try:
                    self._load_routing_table_data(
                        table, app_id, transceiver,
                        routing_table_compressor_app_id, progress_bar, cores,
                        matrix_addresses_and_size[(table.x, table.y)])

                    comms_sdram = transceiver.malloc_sdram(
                        table.x, table.y, SIZE_OF_COMMS_SDRAM,
                        routing_table_compressor_app_id)

                    self._load_address_data(
                        addresses, table.x, table.y, transceiver,
                        routing_table_compressor_app_id,
                        cores, matrix_addresses_and_size[(table.x, table.y)],
                        bit_field_compressor_executable_path,
                        bit_field_sorter_executable_path, comms_sdram,
                        threshold_percentage, retry_count)

                    self._load_usable_sdram(
                        matrix_addresses_and_size[(table.x, table.y)], table.x,
                        table.y, transceiver, routing_table_compressor_app_id,
                        cores)

                    self._load_compressor_data(
                        table.x, table.y, time_per_iteration, transceiver,
                        bit_field_compressor_executable_path, cores,
                        compress_as_much_as_possible, comms_sdram)
                except CantFindSDRAMToUseException:
                    run_by_host.append((table.x, table.y))

        return run_by_host

    def _load_compressor_data(
            self, chip_x, chip_y, time_per_iteration, transceiver,
            bit_field_compressor_executable_path, cores,
            compress_as_much_as_possible, comms_sdram):
        """ Updates the user addresses for the compressor cores with the
            compression settings.

        :param int chip_x: chip x coord
        :param int chip_y: chip y coord
        :param int time_per_iteration: time per attempt of compression
        :param ~.Transceiver transceiver: SpiNNMan instance
        :param str bit_field_compressor_executable_path:
            path for the compressor binary
        :param bool compress_as_much_as_possible:
            whether to compress as much as possible
        :param ExecutableTargets cores: the executable targets
        :param int comms_sdram: Address for comms block
        """
        compressor_cores = cores.get_cores_for_binary(
            bit_field_compressor_executable_path)
        for processor_id in compressor_cores.get_core_subset_for_chip(
                chip_x, chip_y).processor_ids:
            user1_address = \
                transceiver.get_user_1_register_address_from_core(processor_id)
            user2_address = \
                transceiver.get_user_2_register_address_from_core(processor_id)
            user3_address = \
                transceiver.get_user_3_register_address_from_core(processor_id)
            # user 1 the time per compression attempt
            transceiver.write_memory(
                chip_x, chip_y, user1_address,
                int(time_per_iteration * SECOND_TO_MICRO_SECOND))
            # user 2 Compress as much as needed flag
            transceiver.write_memory(
                chip_x, chip_y, user2_address,
                int(compress_as_much_as_possible))
            # user 3 the comms_sdram area
            transceiver.write_memory(
                chip_x, chip_y, user3_address, comms_sdram)

    def _load_usable_sdram(
            self, matrix_addresses_and_size, chip_x, chip_y, transceiver,
            routing_table_compressor_app_id, cores):
        """ loads the addresses of stealable sdram

        :param list(tuple(int,int)) matrix_addresses_and_size:
            SDRAM usable and sizes
        :param int chip_x: the chip x to consider here
        :param int chip_y: the chip y to consider here
        :param ~.Transceiver transceiver: the spinnman instance
        :param int routing_table_compressor_app_id: system app id.
        :param ExecutableTargets cores: the cores that compressor will run on
        """
        address_data = self._generate_chip_matrix_data(
            matrix_addresses_and_size)

        # get sdram address on chip
        try:
            sdram_address = transceiver.malloc_sdram(
                chip_x, chip_y, len(address_data),
                routing_table_compressor_app_id)
        except (SpinnmanInvalidParameterException,
                SpinnmanUnexpectedResponseCodeException):
            sdram_address = self._steal_from_matrix_addresses(
                matrix_addresses_and_size, len(address_data))
            address_data = self._generate_chip_matrix_data(
                matrix_addresses_and_size)

        # write sdram
        transceiver.write_memory(
            chip_x, chip_y, sdram_address, address_data, len(address_data))

        # get the only processor on the chip
        processor_id = list(cores.all_core_subsets.get_core_subset_for_chip(
            chip_x, chip_y).processor_ids)[0]

        # update user 2 with location
        user3_address = transceiver.get_user_3_register_address_from_core(
            processor_id)
        transceiver.write_memory(
            chip_x, chip_y, user3_address, sdram_address)

    def _generate_chip_matrix_data(self, list_of_sizes_and_address):
        """ generate the data for the chip matrix data

        :param list(tuple(int,int)) list_of_sizes_and_address:
            SDRAM addresses and sizes
        :return: byte array of data
        """
        data = b""
        data += self._ONE_WORD.pack(len(list_of_sizes_and_address))
        for (memory_address, size) in list_of_sizes_and_address:
            data += self._TWO_WORDS.pack(memory_address, size)
        return data

    def _load_address_data(
            self, addresses, chip_x, chip_y, transceiver,
            routing_table_compressor_app_id, cores, matrix_addresses_and_size,
            bit_field_compressor_executable_path,
            bit_field_sorter_executable_path, comms_sdram,
            threshold_percentage, retry_count):
        """ loads the bitfield addresses space

        :param dict(tuple(int,int),tuple(int,int)) addresses:
            the addresses to load
        :param int chip_x: the chip x to consider here
        :param int chip_y: the chip y to consider here
        :param ~.Transceiver transceiver: the spinnman instance
        :param int routing_table_compressor_app_id: system app id.
        :param ExecutableTargets cores: the cores that compressor will run on
        :param str bit_field_compressor_executable_path:
            the path to the compressor binary path
        :param str bit_field_sorter_executable_path:
            the path to the sorter binary
        :param int comms_sdram: Address for comms block
        :param int threshold_percentage:
            the percentage of bitfields the user has defined as a minimum
            needed to pass to be successful.
        :param retry_count:
            Number of times that the sorters should set of the compressions
            again. None for as much as needed
        :type retry_count: int or None
        :rtype: None
        """
        # generate address_data
        address_data = self._generate_chip_data(
            addresses[(chip_x, chip_y)],
            cores.get_cores_for_binary(
                bit_field_compressor_executable_path).get_core_subset_for_chip(
                    chip_x, chip_y),
            comms_sdram, threshold_percentage, retry_count)

        # get sdram address on chip
        try:
            sdram_address = transceiver.malloc_sdram(
                chip_x, chip_y, len(address_data),
                routing_table_compressor_app_id)
        except (SpinnmanInvalidParameterException,
                SpinnmanUnexpectedResponseCodeException):
            sdram_address = self._steal_from_matrix_addresses(
                matrix_addresses_and_size, len(address_data))

        # write sdram
        transceiver.write_memory(
            chip_x, chip_y, sdram_address, address_data, len(address_data))

        # get the only processor on the chip
        sorter_cores = cores.get_cores_for_binary(
            bit_field_sorter_executable_path)
        processor_id = list(sorter_cores.get_core_subset_for_chip(
            chip_x, chip_y).processor_ids)[0]

        # update user 2 with location
        user2_address = transceiver.get_user_2_register_address_from_core(
            processor_id)
        transceiver.write_memory(
            chip_x, chip_y, user2_address, sdram_address)

    def _load_routing_table_data(
            self, table, app_id, transceiver,
            routing_table_compressor_app_id, progress_bar, cores,
            matrix_addresses_and_size):
        """ loads the routing table data

        :param AbsractMulticastRoutingTable table: the routing table to load
        :param int app_id: application app id
        :param ~.Transceiver transceiver: spinnman instance
        :param ~.ProgressBar progress_bar: progress bar
        :param int routing_table_compressor_app_id: system app id
        :param ExecutableTargets cores:
            the cores that the compressor going to run on
        :raises CantFindSDRAMToUse: when sdram is not malloc-ed or stolen
        """

        routing_table_data = self._build_routing_table_data(app_id, table)

        # go to spinnman and ask for a memory region of that size per chip.
        try:
            base_address = transceiver.malloc_sdram(
                table.x, table.y, len(routing_table_data),
                routing_table_compressor_app_id)
        except (SpinnmanInvalidParameterException,
                SpinnmanUnexpectedResponseCodeException):
            base_address = self._steal_from_matrix_addresses(
                matrix_addresses_and_size, len(routing_table_data))

        # write SDRAM requirements per chip
        transceiver.write_memory(
            table.x, table.y, base_address, routing_table_data)

        # get the only processor on the chip
        processor_id = list(cores.all_core_subsets.get_core_subset_for_chip(
            table.x, table.y).processor_ids)[0]

        # update user 1 with location
        user1_address = transceiver.get_user_1_register_address_from_core(
            processor_id)
        transceiver.write_memory(
            table.x, table.y, user1_address, base_address)

        # update progress bar
        progress_bar.update()

    def _build_routing_table_data(self, app_id, routing_table):
        """ builds routing data as needed for the compressor cores

        :param int app_id: appid of the application to load entries with
        :param ~.AbsractMulticastRoutingTable routing_table:
            the uncompressed routing table
        :return: data array
        :rtype: bytearray
        """
        data = b''
        data += self._TWO_WORDS.pack(app_id, routing_table.number_of_entries)

        # sort entries based on generality
        sorted_routing_table = sorted(
            routing_table.multicast_routing_entries,
            key=lambda rt_entry: ordered_covering_generality(
                rt_entry.routing_entry_key, rt_entry.mask))

        # write byte array for the sorted table
        for entry in sorted_routing_table:
            data += self._FOUR_WORDS.pack(
                entry.routing_entry_key, entry.mask,
                Router.convert_routing_table_entry_to_spinnaker_route(entry),
                get_defaultable_source_id(entry=entry))
        return bytearray(data)

    @staticmethod
    def _steal_from_matrix_addresses(matrix_addresses_and_size, size_to_steal):
        """ steals memory from synaptic matrix as needed

        :param dict(tuple(int,int),tuple(int,int)) matrix_addresses_and_size:
            matrix addresses and sizes; updated by this method
        :param int size_to_steal: size needed to steal from matrix's.
        :return: address to start steal from
        :rtype: int
        :raises CantFindSDRAMToUseException:
            when no space is big enough to steal from.
        """
        for pos, (base_address, size) in enumerate(matrix_addresses_and_size):
            if size >= size_to_steal:
                new_size = size - size_to_steal
                matrix_addresses_and_size[pos] = (base_address, new_size)
                return base_address
        raise CantFindSDRAMToUseException()

    def _add_to_addresses(
            self, vertex, placement, transceiver, region_addresses,
            sdram_block_addresses_and_sizes):
        """ adds data about the API-based vertex.

        :param AbstractSupportsBitFieldRoutingCompression vertex:
            vertex which utilises the API
        :param ~.Placement placement: placement of vertex
        :param ~.Transceiver transceiver:  spinnman instance
        :param dict(tuple(int,int),list(tuple(int,int))) region_addresses:
            store for data regions
        :param sdram_block_addresses_and_sizes: store for surplus SDRAM
        :type sdram_block_addresses_and_sizes:
            dict(tuple(int,int),list(tuple(int,int)))
        """
        # store the region sdram address's
        bit_field_sdram_address = vertex.bit_field_base_address(
            transceiver, placement)
        region_addresses[placement.x, placement.y].append(
            (bit_field_sdram_address, placement.p))

        # store the available space from the matrix to steal
        blocks = vertex.regeneratable_sdram_blocks_and_sizes(
            transceiver, placement)

        for (address, size) in blocks:
            if size != 0 and size > self._MIN_SIZE_FOR_HEAP:
                sdram_block_addresses_and_sizes[
                    placement.x, placement.y].append((address, size))
        sorted(
            sdram_block_addresses_and_sizes[placement.x, placement.y],
            key=lambda data: data[0])

    def _generate_addresses(
            self, machine_graph, placements, transceiver, progress_bar):
        """ generates the bitfield SDRAM addresses

        :param ~.MachineGraph machine_graph: machine graph
        :param ~.Placements placements: placements
        :param ~.Transceiver transceiver: spinnman instance
        :param ~.ProgressBar progress_bar: the progress bar
        :return: region_addresses and the executable targets to load the
            router table compressor with bitfield. and the SDRAM blocks
            available for use on each core that we plan to use
        :rtype: tuple(dict(tuple(int,int),tuple(int,int)),
            dict(tuple(int,int),list(tuple(int,int))))
        """
        # data holders
        region_addresses = defaultdict(list)
        sdram_block_addresses_and_sizes = defaultdict(list)

        for vertex in progress_bar.over(
                machine_graph.vertices, finish_at_end=False):
            placement = placements.get_placement_of_vertex(vertex)

            # locate the interface vertex (maybe app or machine)
            if isinstance(
                    vertex, AbstractSupportsBitFieldRoutingCompression):
                self._add_to_addresses(
                    vertex, placement, transceiver, region_addresses,
                    sdram_block_addresses_and_sizes)

        return region_addresses, sdram_block_addresses_and_sizes

    def _generate_chip_data(
            self, address_list, cores, comms_sdram, threshold_percentage,
            retry_count):
        """
        Generate the region_addresses_t data

        Minimum percentage of bitfields to be merge in (currently ignored)

        Number of times that the sorters should set of the compressions again

        Pointer to the area malloced to hold the comms_sdram

        Number of processors in the list

        The data for the processors

        :param list(tuple(int,int)) address_list:
            the list of SDRAM addresses
        :param ~.CoreSubset cores: compressor cores on this chip.
        :param int comms_sdram: Address for comms block
        :param int threshold_percentage:
            the percentage of bitfields the user has defined as a minimum
            needed to pass to be successful.
        :param retry_count:
            Number of times that the sorters should set of the compressions
            again. None for as much as needed
        :type retry_count: int or None
        :return: the byte array
        :rtype: bytes
        """
        data = self._FOUR_WORDS.pack(
            threshold_percentage,
            retry_count if retry_count is not None else 0xFFFFFFFF,
            comms_sdram, len(address_list))
        for (bit_field, processor_id) in address_list:
            data += self._TWO_WORDS.pack(bit_field, processor_id)
        data += self._ONE_WORD.pack(len(cores))
        data += n_word_struct(len(cores)).pack(*list(cores.processor_ids))
        return data


class MachineBitFieldOrderedCoveringCompressor(
        MachineBitFieldRouterCompressor):

    @property
    @overrides(MachineBitFieldRouterCompressor.compressor_aplx)
    def compressor_aplx(self):
        return "bit_field_ordered_covering_compressor.aplx"

    @property
    @overrides(MachineBitFieldRouterCompressor.compressor_type)
    def compressor_type(self):
        return "OrderedCovering"


class MachineBitFieldUnorderedRouterCompressor(
        MachineBitFieldRouterCompressor):
    """ DEPRACATED use MachineBitFieldOrderedCoveringCompressor """

    def __new__(cls, *args, **kwargs):
        logger.warning(
            "MachineBitFieldUnorderedRouterCompressor algorithm name is "
            "deprecated. "
            "Please use MachineBitFieldOrderedCoveringCompressor instead. "
            "loading_algorithms from your cfg to use defaults")
        return super().__new__(cls, *args, **kwargs)


class MachineBitFieldPairRouterCompressor(MachineBitFieldRouterCompressor):

    @property
    @overrides(MachineBitFieldRouterCompressor.compressor_aplx)
    def compressor_aplx(self):
        return "bit_field_pair_compressor.aplx"

    @property
    @overrides(MachineBitFieldRouterCompressor.compressor_type)
    def compressor_type(self):
        return "Pair"



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict
import logging
import numpy
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.log import FormatAdapter
from spinn_machine import CoreSubsets
from data_specification import DataSpecificationExecutor
from data_specification.constants import MAX_MEM_REGIONS
from data_specification.exceptions import DataSpecificationException
from spinn_front_end_common.interface.ds.ds_write_info import DsWriteInfo
from spinn_front_end_common.utilities.helpful_functions import (
    write_address_to_user0)
from spinn_front_end_common.utilities.utility_objs import (
    ExecutableType, DataWritten)
from spinn_front_end_common.utilities.emergency_recovery import (
    emergency_recover_states_from_failure)

logger = FormatAdapter(logging.getLogger(__name__))
_MEM_REGIONS = range(MAX_MEM_REGIONS)


def system_cores(exec_targets):
    """ Get the subset of cores that are to be used for system operations.

    :param ~spinnman.model.ExecutableTargets exec_targets:
    :rtype: ~spinn_machine.CoreSubsets
    """
    cores = CoreSubsets()
    for binary in exec_targets.get_binaries_of_executable_type(
            ExecutableType.SYSTEM):
        cores.add_core_subsets(exec_targets.get_cores_for_binary(binary))
    return cores


def filter_out_system_executables(dsg_targets, executable_targets):
    """ Select just the application DSG loading tasks

    :param DataSpecificationTargets dsg_tagets:
    :param ~spinnman.model.ExecutableTargets executable_targets:
    :rtype: dict(tuple(int,int,int), ~io.RawIOBase)
    """
    syscores = system_cores(executable_targets)
    return OrderedDict(
        (core, spec) for (core, spec) in dsg_targets.items()
        if core not in syscores)


def filter_out_app_executables(dsg_targets, executable_targets):
    """ Select just the system DSG loading tasks

    :param DataSpecificationTargets dsg_tagets:
    :param ~spinnman.model.ExecutableTargets executable_targets:
    :rtype: dict(tuple(int,int,int), ~io.RawIOBase)
    """
    syscores = system_cores(executable_targets)
    return OrderedDict(
        (core, spec) for (core, spec) in dsg_targets.items()
        if core in syscores)


class HostExecuteDataSpecification(object):
    """ Executes the host based data specification.
    """

    __slots__ = [
        # the application ID of the simulation
        "_app_id",
        "_core_to_conn_map",
        # The path where the SQLite database holding the data will be placed,
        # and where any java provenance can be written.
        "_db_folder",
        # The support class to run via Java. If None pure python is used.
        "_java",
        # The python representation of the SpiNNaker machine.
        "_machine",
        "_monitors",
        "_placements",
        # The spinnman instance.
        "_txrx",
        # The write info; a dict of cores to a dict of
        # 'start_address', 'memory_used', 'memory_written'
        "_write_info_map"]

    first = True

    def __init__(self):
        self._app_id = None
        self._core_to_conn_map = None
        self._db_folder = None
        self._java = None
        self._machine = None
        self._monitors = None
        self._placements = None
        self._txrx = None
        self._write_info_map = None

    def __java_database(self, dsg_targets, progress, region_sizes):
        """
        :param DataSpecificationTargets dsg_tagets:
        :param ~spinn_utilities.progress_bar.ProgressBar progress:
        :param dict(tuple(int,int,int)int) region_sizes:
        :rtype: DsWriteInfo
        """
        # Copy data from WriteMemoryIOData to database
        dw_write_info = DsWriteInfo(dsg_targets.get_database())
        dw_write_info.clear_write_info()
        if self._write_info_map is not None:
            for core, info in self._write_info_map.items():
                (x, y, p) = core
                dw_write_info.set_info(x, y, p, info)
                del region_sizes[core]
        for core in region_sizes:
            (x, y, p) = core
            dw_write_info.set_size_info(x, y, p, region_sizes[core])
        progress.update()
        dsg_targets.set_app_id(self._app_id)
        self._java.set_machine(self._machine)
        self._java.set_report_folder(self._db_folder)
        progress.update()
        return dw_write_info

    def __java_all(self, dsg_targets, region_sizes):
        """ Does the Data Specification Execution and loading using Java

        :param DataSpecificationTargets dsg_targets:
            map of placement to file path
        :return: map of of cores to descriptions of what was written
        :rtype: DsWriteInfo
        """
        # create a progress bar for end users
        progress = ProgressBar(
            3, "Executing data specifications and loading data using Java")

        # Copy data from WriteMemoryIOData to database
        dw_write_info = self.__java_database(
            dsg_targets, progress, region_sizes)

        self._java.execute_data_specification()

        progress.end()
        return dw_write_info

    def __python_all(self, dsg_targets, region_sizes):
        """ Does the Data Specification Execution and loading using Python

        :param DataSpecificationTargets dsg_targets:
            map of placement to file path
        :param dict(tuple(int,int,int),int) region_sizes:
            map between vertex and list of region sizes
        :return: dict of cores to descriptions of what was written
        :rtype: dict(tuple(int,int,int), DataWritten)
        """
        # While the database supports having the info in it a python bugs does
        # not like iterating over and writing intermingled so using a dict
        results = self._write_info_map
        if results is None:
            results = dict()

        # create a progress bar for end users
        progress = ProgressBar(
            dsg_targets.n_targets(),
            "Executing data specifications and loading data")

        # allocate and set user 0 before loading data
        base_addresses = dict()
        for core, _ in dsg_targets.items():
            base_addresses[core] = self.__malloc_region_storage(
                core, region_sizes[core])

        for core, reader in progress.over(dsg_targets.items()):
            results[core] = self.__python_execute(
                core, reader, self._txrx.write_memory,
                base_addresses[core], region_sizes[core])

        return results

    def execute_application_data_specs(
            self, transceiver, machine, app_id, dsg_targets,
            uses_advanced_monitors, executable_targets, region_sizes,
            placements=None, extra_monitor_cores=None,
            extra_monitor_cores_to_ethernet_connection_map=None,
            report_folder=None, java_caller=None,
            processor_to_app_data_base_address=None,
            disable_advanced_monitor_usage=False):
        """ Execute the data specs for all non-system targets.

        :param ~spinn_machine.Machine machine:
            the python representation of the SpiNNaker machine
        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnman instance
        :param int app_id: the application ID of the simulation
        :param dict(tuple(int,int,int),int) region_sizes:
            the coord for region sizes for each core
        :param DataSpecificationTargets dsg_targets:
            map of placement to file path
        :param bool uses_advanced_monitors:
            whether to use fast data in protocol
        :param ~spinnman.model.ExecutableTargets executable_targets:
            what core will running what binary
        :param ~pacman.model.placements.Placements placements:
            where vertices are located
        :param list(ExtraMonitorSupportMachineVertex) extra_monitor_cores:
            the deployed extra monitors, if any
        :param extra_monitor_cores_to_ethernet_connection_map:
            how to talk to extra monitor cores
        :type extra_monitor_cores_to_ethernet_connection_map:
            dict(tuple(int,int), DataSpeedUpPacketGatherMachineVertex)
        :param processor_to_app_data_base_address:
            map of placement and DSG data
        :type processor_to_app_data_base_address:
            dict(tuple(int,int,int), DsWriteInfo)
        :param bool disable_advanced_monitor_usage:
            whether to avoid using advanced monitors even if they're available
        :return: map of placement and DSG data
        :rtype: dict(tuple(int,int,int),DataWritten) or DsWriteInfo
        """
        # pylint: disable=too-many-arguments
        if processor_to_app_data_base_address is None:
            processor_to_app_data_base_address = dict()
        self._write_info_map = processor_to_app_data_base_address
        self._db_folder = report_folder
        self._java = java_caller
        self._machine = machine
        self._txrx = transceiver
        self._app_id = app_id
        self._monitors = extra_monitor_cores
        self._placements = placements
        self._core_to_conn_map = extra_monitor_cores_to_ethernet_connection_map

        # Allow override to disable
        if disable_advanced_monitor_usage:
            uses_advanced_monitors = False

        impl_method = self.__java_app if java_caller else self.__python_app
        try:
            return impl_method(
                dsg_targets, executable_targets, uses_advanced_monitors,
                region_sizes)
        except:  # noqa: E722
            if uses_advanced_monitors:
                emergency_recover_states_from_failure(
                    self._txrx, self._app_id, executable_targets)
            raise

    def __set_router_timeouts(self):
        for receiver in self._core_to_conn_map.values():
            receiver.load_system_routing_tables(
                self._txrx, self._monitors, self._placements)
            receiver.set_cores_for_data_streaming(
                self._txrx, self._monitors, self._placements)

    def __reset_router_timeouts(self):
        # reset router timeouts
        for receiver in self._core_to_conn_map.values():
            receiver.unset_cores_for_data_streaming(
                self._txrx, self._monitors, self._placements)
            # reset router tables
            receiver.load_application_routing_tables(
                self._txrx, self._monitors, self._placements)

    def __select_writer(self, x, y):
        chip = self._machine.get_chip_at(x, y)
        ethernet_chip = self._machine.get_chip_at(
            chip.nearest_ethernet_x, chip.nearest_ethernet_y)
        gatherer = self._core_to_conn_map[ethernet_chip.x, ethernet_chip.y]
        return gatherer.send_data_into_spinnaker

    def __python_app(
            self, dsg_targets, executable_targets, use_monitors,
            region_sizes):
        """
        :param DataSpecificationTargets dsg_targets:
        :param ~spinnman.model.ExecutableTargets executable_targets:
        :param bool use_monitors:
        :param dict(tuple(int,int,int),int) region_sizes:
        :return: dict of cores to descriptions of what was written
        :rtype: dict(tuple(int,int,int),DataWritten)
        """
        dsg_targets = filter_out_system_executables(
            dsg_targets, executable_targets)

        if use_monitors:
            self.__set_router_timeouts()

        # create a progress bar for end users
        progress = ProgressBar(
            len(dsg_targets) * 2,
            "Executing data specifications and loading data for "
            "application vertices")

        # allocate and set user 0 before loading data
        base_addresses = dict()
        for core, _ in progress.over(dsg_targets.items(), finish_at_end=False):
            base_addresses[core] = self.__malloc_region_storage(
                core, region_sizes[core])

        for core, reader in progress.over(dsg_targets.items()):
            x, y, _p = core
            # write information for the memory map report
            self._write_info_map[core] = self.__python_execute(
                core, reader,
                self.__select_writer(x, y)
                if use_monitors else self._txrx.write_memory,
                base_addresses[core], region_sizes[core])

        if use_monitors:
            self.__reset_router_timeouts()
        return self._write_info_map

    def __java_app(
            self, dsg_targets, executable_targets, use_monitors,
            region_sizes):
        """
        :param DataSpecificationTargets dsg_targets:
        :param ~spinnman.model.ExecutableTargets executable_targets:
        :param bool use_monitors:
        :param dict(tuple(int,int,int),int) region_sizes:
        :return: map of cores to descriptions of what was written
        :rtype: DsWriteInfo
        """
        # create a progress bar for end users
        progress = ProgressBar(
            4, "Executing data specifications and loading data for "
            "application vertices using Java")

        dsg_targets.mark_system_cores(system_cores(executable_targets))
        progress.update()

        # Copy data from WriteMemoryIOData to database
        dw_write_info = self.__java_database(
            dsg_targets, progress, region_sizes)
        if use_monitors:
            self._java.set_placements(self._placements, self._txrx)

        self._java.execute_app_data_specification(use_monitors)

        progress.end()
        return dw_write_info

    def execute_system_data_specs(
            self, transceiver, machine, app_id, dsg_targets, region_sizes,
            executable_targets, report_folder=None,
            java_caller=None, processor_to_app_data_base_address=None):
        """ Execute the data specs for all system targets.

        :param ~spinnman.transceiver.Transceiver transceiver:
            the spinnman instance
        :param ~spinn_machine.Machine machine:
            the python representation of the spinnaker machine
        :param int app_id: the application ID of the simulation
        :param dict(tuple(int,int,int),str) dsg_targets:
            map of placement to file path
        :param dict(tuple(int,int,int),int) region_sizes:
            the coordinates for region sizes for each core
        :param ~spinnman.model.ExecutableTargets executable_targets:
            the map between binaries and locations and executable types
        :param str report_folder:
        :param JavaCaller java_caller:
        :param processor_to_app_data_base_address:
        :type processor_to_app_data_base_address:
            dict(tuple(int,int,int),DataWritten)
        :return: map of placement and DSG data, and loaded data flag.
        :rtype: dict(tuple(int,int,int),DataWritten) or DsWriteInfo
        """
        # pylint: disable=too-many-arguments

        if processor_to_app_data_base_address is None:
            processor_to_app_data_base_address = dict()
        self._write_info_map = processor_to_app_data_base_address
        self._machine = machine
        self._txrx = transceiver
        self._app_id = app_id
        self._db_folder = report_folder
        self._java = java_caller
        impl_method = self.__java_sys if java_caller else self.__python_sys
        return impl_method(dsg_targets, executable_targets, region_sizes)

    def __java_sys(self, dsg_targets, executable_targets, region_sizes):
        """ Does the Data Specification Execution and loading using Java

        :param DataSpecificationTargets dsg_targets:
            map of placement to file path
        :param ~spinnman.model.ExecutableTargets executable_targets:
            the map between binaries and locations and executable types
        :param dict(tuple(int,int,int),int) region_sizes:
            the coord for region sizes for each core
        :return: map of cores to descriptions of what was written
        :rtype: DsWriteInfo
        """
        # create a progress bar for end users
        progress = ProgressBar(
            4, "Executing data specifications and loading data for system "
            "vertices using Java")

        dsg_targets.mark_system_cores(system_cores(executable_targets))
        progress.update()

        # Copy data from WriteMemoryIOData to database
        dw_write_info = self.__java_database(
            dsg_targets, progress, region_sizes)

        self._java.execute_system_data_specification()

        progress.end()
        return dw_write_info

    def __python_sys(self, dsg_targets, executable_targets, region_sizes):
        """ Does the Data Specification Execution and loading using Python

        :param DataSpecificationTargets dsg_targets:
            map of placement to file path
        :param ~spinnman.model.ExecutableTargets executable_targets:
            the map between binaries and locations and executable types
        :param dict(tuple(int,int,int),int) region_sizes:
            the coord for region sizes for each core
        :return: dict of cores to descriptions of what was written
        :rtype: dict(tuple(int,int,int),DataWritten)
        """
        # While the database supports having the info in it a python bugs does
        # not like iterating over and writing intermingled so using a dict
        sys_targets = filter_out_app_executables(
            dsg_targets, executable_targets)

        # create a progress bar for end users
        progress = ProgressBar(
            len(sys_targets) * 2,
            "Executing data specifications and loading data for "
            "system vertices")

        # allocate and set user 0 before loading data
        base_addresses = dict()
        for core, _ in progress.over(sys_targets.items(), finish_at_end=False):
            base_addresses[core] = self.__malloc_region_storage(
                core, region_sizes[core])

        for core, reader in progress.over(sys_targets.items()):
            self._write_info_map[core] = self.__python_execute(
                core, reader, self._txrx.write_memory, base_addresses[core],
                region_sizes[core])

        return self._write_info_map

    def __malloc_region_storage(self, core, size):
        """ Allocates the storage for all DSG regions on the core and tells \
            the core and our caller where that storage is.

        :param tuple(int,int,int) core: Which core we're talking about.
        :param int size:
            The total size of all storage for regions on that core, including
            for the header metadata.
        :return: address of region header table (not yet filled)
        :rtype: int
        """
        (x, y, p) = core

        # allocate memory where the app data is going to be written; this
        # raises an exception in case there is not enough SDRAM to allocate
        start_address = self._txrx.malloc_sdram(x, y, size, self._app_id)

        # set user 0 register appropriately to the application data
        write_address_to_user0(self._txrx, x, y, p, start_address)

        return start_address

    def __python_execute(
            self, core, reader, writer_func, base_address, size_allocated):
        """
        :param tuple(int,int,int) core:
        :param ~.AbstractDataReader reader:
        :param callable(tuple(int,int,int,bytearray),None) writer_func:
        :param int base_address:
        :param int size_allocated:
        :rtype: DataWritten
        """
        x, y, p = core

        # Maximum available memory.
        # However, system updates the memory available independently, so the
        # space available check actually happens when memory is allocated.
        memory_available = self._machine.get_chip_at(x, y).sdram.size

        # generate data spec executor
        executor = DataSpecificationExecutor(reader, memory_available)

        # run data spec executor
        try:
            executor.execute()
        except DataSpecificationException:
            logger.error("Error executing data specification for {}, {}, {}",
                         x, y, p)
            raise

        # Do the actual writing ------------------------------------

        # Write the header and pointer table
        header = executor.get_header()
        pointer_table = executor.get_pointer_table(base_address)
        data_to_write = numpy.concatenate((header, pointer_table)).tostring()

        # NB: DSE meta-block is always small (i.e., one SDP write)
        self._txrx.write_memory(x, y, base_address, data_to_write)
        bytes_written = len(data_to_write)

        # Write each region
        for region_id in _MEM_REGIONS:
            region = executor.get_region(region_id)
            if region is None:
                continue
            max_pointer = region.max_write_pointer
            if region.unfilled or max_pointer == 0:
                continue

            # Get the data up to what has been written
            data = region.region_data[:max_pointer]

            # Write the data to the position
            writer_func(x, y, pointer_table[region_id], data)
            bytes_written += len(data)

        return DataWritten(base_address, size_allocated, bytes_written)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
from .dpri_flags import DPRIFlags

_PATTERN = struct.Struct("<IIIIIIIII")


def _decode_router_timeout_value(value):
    """ Get the timeout value of a router in ticks, given an 8-bit floating\
        point value stored in an int (!)

    :param int value: The value to convert
    :rtype: int
    """
    mantissa = value & 0xF
    exponent = (value >> 4) & 0xF
    if exponent <= 4:
        return ((mantissa + 16) - (2 ** (4 - exponent))) * (2 ** exponent)
    return (mantissa + 16) * (2 ** exponent)


class ReInjectionStatus(object):
    """ Represents a status information report from dropped packet reinjection.
    """

    __slots__ = (

        # The WAIT1 timeout value of the router in cycles
        "_wait1_timeout",

        # The WAIT2 timeout value of the router in cycles
        "_wait2_timeout",

        # The number of packets dropped by the router and received by\
        # the re injection functionality (may not fit in the queue though)
        "_n_dropped_packets",

        # The number of times that when a dropped packet was read it was\
        #    found that another one or more packets had also been dropped,\
        #    but had been missed
        "_n_missed_dropped_packets",

        # Of the n_dropped_packets received, how many were lost due to not\
        #    having enough space in the queue of packets to reinject
        "_n_dropped_packet_overflows",

        # Of the n_dropped_packets received, how many packets were\
        #    successfully re-injected
        "_n_reinjected_packets",

        # The number of times that when a dropped packet was caused due to\
        # a link failing to take the packet.
        "_n_link_dumps",

        # The number of times that when a dropped packet was caused due to\
        # a processor failing to take the packet.
        "_n_processor_dumps",

        # the flags that states which types of packets were being recorded
        "_flags"
    )

    def __init__(self, data, offset):
        """
        :param bytes data: The data containing the information
        :param int offset: The offset in the data where the information starts
        """
        (self._wait1_timeout, self._wait2_timeout,
         self._n_dropped_packets, self._n_missed_dropped_packets,
         self._n_dropped_packet_overflows, self._n_reinjected_packets,
         self._n_link_dumps, self._n_processor_dumps, self._flags) = \
            _PATTERN.unpack_from(data, offset)

    @property
    def router_wait1_timeout(self):
        """ The WAIT1 timeout value of the router, in cycles.

        :rtype: int
        """
        return _decode_router_timeout_value(self._wait1_timeout)

    @property
    def router_wait1_timeout_parameters(self):
        """ The WAIT1 timeout value of the router as mantissa and exponent.

        :rtype: tuple(int,int)
        """
        mantissa = self._wait1_timeout & 0xF
        exponent = (self._wait1_timeout >> 4) & 0xF
        return mantissa, exponent

    @property
    def router_wait2_timeout(self):
        """ The WAIT2 timeout value of the router, in cycles.

        :rtype: int
        """
        return _decode_router_timeout_value(self._wait2_timeout)

    @property
    def router_wait2_timeout_parameters(self):
        """ The WAIT2 timeout value of the router as mantissa and exponent.

        :rtype: tuple(int,int)
        """
        mantissa = self._wait2_timeout & 0xF
        exponent = (self._wait2_timeout >> 4) & 0xF
        return mantissa, exponent

    @property
    def n_dropped_packets(self):
        """ The number of packets dropped by the router and received by\
            the reinjection functionality (may not fit in the queue though)

        :rtype: int
        """
        return self._n_dropped_packets

    @property
    def n_missed_dropped_packets(self):
        """ The number of times that when a dropped packet was read it was\
            found that another one or more packets had also been dropped,\
            but had been missed

        :rtype: int
        """
        return self._n_missed_dropped_packets

    @property
    def n_dropped_packet_overflows(self):
        """ Of the n_dropped_packets received, how many were lost due to not\
            having enough space in the queue of packets to reinject

        :rtype: int
        """
        return self._n_dropped_packet_overflows

    @property
    def n_processor_dumps(self):
        """ The number of times that when a dropped packet was caused due to\
            a processor failing to take the packet.

        :rtype: int
        """
        return self._n_processor_dumps

    @property
    def n_link_dumps(self):
        """ The number of times that when a dropped packet was caused due to\
            a link failing to take the packet.

        :rtype: int
        """
        return self._n_link_dumps

    @property
    def n_reinjected_packets(self):
        """ Of the n_dropped_packets received, how many packets were\
            successfully re injected

        :rtype: int
        """
        return self._n_reinjected_packets

    def _flag_set(self, flag):
        return (self._flags & flag.value) != 0

    @property
    def is_reinjecting_multicast(self):
        """ True if re-injection of multicast packets is enabled

        :rtype: bool
        """
        return self._flag_set(DPRIFlags.MULTICAST)

    @property
    def is_reinjecting_point_to_point(self):
        """ True if re-injection of point-to-point packets is enabled

        :rtype: bool
        """
        return self._flag_set(DPRIFlags.POINT_TO_POINT)

    @property
    def is_reinjecting_nearest_neighbour(self):
        """ True if re-injection of nearest neighbour packets is enabled

        :rtype: bool
        """
        return self._flag_set(DPRIFlags.NEAREST_NEIGHBOUR)

    @property
    def is_reinjecting_fixed_route(self):
        """ True if re-injection of fixed-route packets is enabled

        :rtype: bool
        """
        return self._flag_set(DPRIFlags.FIXED_ROUTE)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
The code in this module is intended primarily for being invoked via the
PACMAN Executor.
"""

import os
from .application_finisher import ApplicationFinisher
from .application_runner import ApplicationRunner
from .buffer_extractor import BufferExtractor
from .chip_iobuf_clearer import ChipIOBufClearer
from .chip_iobuf_extractor import ChipIOBufExtractor
from .buffer_manager_creator import BufferManagerCreator
from .chip_provenance_updater import ChipProvenanceUpdater
from .chip_runtime_updater import ChipRuntimeUpdater
from .compute_energy_used import ComputeEnergyUsed
from .database_interface import DatabaseInterface
from .system_multicast_routing_generator import (
    SystemMulticastRoutingGenerator)
from .dsg_region_reloader import DSGRegionReloader
from .edge_to_n_keys_mapper import EdgeToNKeysMapper
from .energy_provenance_reporter import EnergyProvenanceReporter
from .finalise_timing_data import FinaliseTimingData
from .find_application_chips_used import FindApplicationChipsUsed
from .graph_binary_gatherer import GraphBinaryGatherer
from .graph_data_specification_writer import (
    GraphDataSpecificationWriter)
from .graph_measurer import GraphMeasurer
from .graph_provenance_gatherer import GraphProvenanceGatherer
from .hbp_allocator import HBPAllocator
from .hbp_max_machine_generator import HBPMaxMachineGenerator
from .host_bit_field_router_compressor import HostBasedBitFieldRouterCompressor
from .host_execute_data_specification import HostExecuteDataSpecification
from .insert_chip_power_monitors_to_graphs import (
    InsertChipPowerMonitorsToGraphs)
from .insert_edges_to_extra_monitor_functionality import (
    InsertEdgesToExtraMonitorFunctionality)
from .insert_edges_to_live_packet_gatherers import (
    InsertEdgesToLivePacketGatherers)
from .insert_extra_monitor_vertices_to_graphs import (
    InsertExtraMonitorVerticesToGraphs)
from .insert_live_packet_gatherers_to_graphs import (
    InsertLivePacketGatherersToGraphs)
from .load_executable_images import LoadExecutableImages
from .load_fixed_routes import LoadFixedRoutes
from .local_tdma_builder import LocalTDMABuilder
from .locate_executable_start_type import LocateExecutableStartType
from .machine_bit_field_router_compressor import (
    MachineBitFieldRouterCompressor)
from .machine_generator import MachineGenerator
from .create_notification_protocol import CreateNotificationProtocol
from .placements_provenance_gatherer import PlacementsProvenanceGatherer
from .pre_allocate_for_bit_field_router_compressor import (
    PreAllocateForBitFieldRouterCompressor)
from .pre_allocate_resources_for_chip_power_monitor import (
    PreAllocateResourcesForChipPowerMonitor)
from .pre_allocate_resources_for_live_packet_gatherers import (
    PreAllocateResourcesForLivePacketGatherers)
from .preallocate_resources_for_extra_monitor_support import (
    PreAllocateResourcesForExtraMonitorSupport)
from .process_partition_constraints import ProcessPartitionConstraints
from .profile_data_gatherer import ProfileDataGatherer
from .provenance_json_writer import ProvenanceJSONWriter
from .provenance_sql_writer import ProvenanceSQLWriter
from .provenance_xml_writer import ProvenanceXMLWriter
from .router_provenance_gatherer import RouterProvenanceGatherer
from .routing_setup import RoutingSetup
from .routing_table_loader import RoutingTableLoader
from .spalloc_allocator import SpallocAllocator
from .spalloc_max_machine_generator import SpallocMaxMachineGenerator
from .tags_loader import TagsLoader
from .tdma_agenda_builder import TDMAAgendaBuilder
from .virtual_machine_generator import VirtualMachineGenerator
from .read_routing_tables_from_machine import ReadRoutingTablesFromMachine
from .sdram_outgoing_partition_allocator import SDRAMOutgoingPartitionAllocator


def interface_xml():
    return os.path.join(
        os.path.dirname(__file__), "front_end_common_interface_functions.xml")


__all__ = [
    "ApplicationFinisher",
    "ApplicationRunner", "BufferExtractor",
    "BufferManagerCreator", "ChipIOBufClearer",
    "ChipIOBufExtractor", "ChipProvenanceUpdater",
    "ChipRuntimeUpdater", "CreateNotificationProtocol",
    "ComputeEnergyUsed", "DatabaseInterface",
    "SystemMulticastRoutingGenerator",
    "DSGRegionReloader", "EdgeToNKeysMapper",
    "EnergyProvenanceReporter", "FinaliseTimingData",
    "FindApplicationChipsUsed",
    "GraphBinaryGatherer", "GraphDataSpecificationWriter",
    "GraphMeasurer", "GraphProvenanceGatherer",
    "HBPAllocator", "HostBasedBitFieldRouterCompressor",
    "HBPMaxMachineGenerator",
    "HostExecuteDataSpecification",
    "InsertChipPowerMonitorsToGraphs",
    "InsertEdgesToExtraMonitorFunctionality",
    "InsertEdgesToLivePacketGatherers",
    "InsertExtraMonitorVerticesToGraphs",
    "InsertLivePacketGatherersToGraphs", "interface_xml",
    "LoadExecutableImages", "LoadFixedRoutes", "LocalTDMABuilder",
    "LocateExecutableStartType", "MachineBitFieldRouterCompressor",
    "MachineGenerator", "PlacementsProvenanceGatherer",
    "PreAllocateForBitFieldRouterCompressor",
    "PreAllocateResourcesForChipPowerMonitor",
    "PreAllocateResourcesForExtraMonitorSupport",
    "PreAllocateResourcesForLivePacketGatherers",
    "ProcessPartitionConstraints", "ProfileDataGatherer",
    "ProvenanceJSONWriter", "ProvenanceSQLWriter", "ProvenanceXMLWriter",
    "ReadRoutingTablesFromMachine", "RouterProvenanceGatherer", "RoutingSetup",
    "RoutingTableLoader", "SDRAMOutgoingPartitionAllocator",
    "SpallocAllocator", "SpallocMaxMachineGenerator", "TagsLoader",
    "TDMAAgendaBuilder", "VirtualMachineGenerator"]



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar


class RoutingTableLoader(object):
    """ Loads routes into initialised routers.
    """
    __slots__ = []

    def __call__(self, router_tables, app_id, transceiver, machine):
        """
        :param router_tables:
        :type router_tables:
            ~pacman.model.routing_tables.MulticastRoutingTables
        :param int app_id:
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~spinn_machine.Machine machine:
        """
        progress = ProgressBar(router_tables.routing_tables,
                               "Loading routing data onto the machine")

        # load each router table that is needed for the application to run into
        # the chips SDRAM
        for table in progress.over(router_tables.routing_tables):
            if (not machine.get_chip_at(table.x, table.y).virtual
                    and table.multicast_routing_entries):
                transceiver.load_multicast_routes(
                    table.x, table.y, table.multicast_routing_entries,
                    app_id=app_id)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import requests
from spinn_utilities.overrides import overrides
from spinn_front_end_common.abstract_models.impl import (
    MachineAllocationController)
from spinn_front_end_common.abstract_models import (
    AbstractMachineAllocationController)
from pacman.exceptions import PacmanConfigurationException


class _HBPJobController(MachineAllocationController):
    __slots__ = [
        # the URLs to call the HBP system
        "_extend_lease_url",
        "_check_lease_url",
        "_release_machine_url",
        "_set_power_url",
        "_where_is_url",
        "_machine_name",
        "_power_on"
    ]

    _WAIT_TIME_MS = 10000

    def __init__(self, url, machine_name):
        """
        :param str url:
        :param str machine_name:
        """
        self._extend_lease_url = "{}/extendLease".format(url)
        self._check_lease_url = "{}/checkLease".format(url)
        self._release_machine_url = url
        self._set_power_url = "{}/power".format(url)
        self._where_is_url = "{}/chipCoordinates".format(url)
        self._machine_name = machine_name
        self._power_on = True
        # Lower the level of requests to WARNING to avoid extra messages
        logging.getLogger("requests").setLevel(logging.WARNING)
        super().__init__("HBPJobController")

    @overrides(AbstractMachineAllocationController.extend_allocation)
    def extend_allocation(self, new_total_run_time):
        r = requests.get(self._extend_lease_url, params={
            "runTime": new_total_run_time})
        r.raise_for_status()

    def _check_lease(self, wait_time):
        r = requests.get(self._check_lease_url, params={
            "waitTime": wait_time})
        r.raise_for_status()
        return r.json()

    def _release(self, machine_name):
        r = requests.delete(self._release_machine_url, params={
            "machineName": machine_name})
        r.raise_for_status()

    def _set_power(self, machine_name, power_on):
        r = requests.put(self._set_power_url, params={
            "machineName": machine_name, "on": bool(power_on)})
        r.raise_for_status()

    def _where_is(self, machine_name, chip_x, chip_y):
        r = requests.get(self._where_is_url, params={
            "machineName": machine_name, "chipX": chip_x,
            "chipY": chip_y})
        r.raise_for_status()
        return r.json()

    @overrides(AbstractMachineAllocationController.close)
    def close(self):
        super().close()
        self._release(self._machine_name)

    @overrides(MachineAllocationController._teardown)
    def _teardown(self):
        self._release(self._machine_name)

    @property
    def power(self):
        return self._power_on

    def set_power(self, power):
        self._set_power(self._machine_name, power)
        self._power_on = power

    @overrides(AbstractMachineAllocationController.where_is_machine)
    def where_is_machine(self, chip_x, chip_y):
        return self._where_is(self._machine_name, chip_x, chip_y)

    @overrides(MachineAllocationController._wait)
    def _wait(self):
        return self._check_lease(self._WAIT_TIME_MS)["allocated"]


class HBPAllocator(object):
    """ Request a machine from the HBP remote access server that will fit\
        a number of chips.
    """

    def __call__(
            self, hbp_server_url, total_run_time, n_chips=None, n_boards=None):
        """
        :param str hbp_server_url:
            The URL of the HBP server from which to get the machine
        :param int total_run_time: The total run time to request
        :param int n_chips: The number of chips required.
            Only used if n_boards is None
        :param int n_boards: The number of boards required
        :return: machine name, machine version, BMP details (if any),
            reset on startup flag, auto-detect BMP, SCAMP connection details,
            boot port, allocation controller
        :rtype: tuple(str, int, object, bool, bool, object, object,
            MachineAllocationController)
        :raises ~pacman.exceptions.PacmanConfigurationException:
            If neither `n_chips` or `n_boards` provided
        """

        url = hbp_server_url
        if url.endswith("/"):
            url = url[:-1]

        machine = self._get_machine(url, n_chips, n_boards, total_run_time)
        hbp_job_controller = _HBPJobController(url, machine["machineName"])

        bmp_details = None
        if "bmp_details" in machine:
            bmp_details = machine["bmpDetails"]

        return (
            machine["machineName"], int(machine["version"]),
            bmp_details, False, False, None, None,
            hbp_job_controller)

    def _get_machine(self, url, n_chips, n_boards, total_run_time):
        """
        :param str url:
        :param int n_chips:
        :param int n_boards:
        :param int total_run_time:
        :rtype: dict
        """
        if n_boards:
            get_machine_request = requests.get(
                url, params={"nBoards": n_boards, "runTime": total_run_time})
        elif n_chips:
            get_machine_request = requests.get(
                url, params={"nChips": n_chips, "runTime": total_run_time})
        else:
            raise PacmanConfigurationException(
                "At least one of n_chips or n_boards must be provided")

        get_machine_request.raise_for_status()
        return get_machine_request.json()



# Copyright (c) 2019-2020 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from pacman.model.resources import (
    SpecificChipSDRAMResource, PreAllocatedResourceContainer)
from spinn_front_end_common.interface.interface_functions. \
    machine_bit_field_router_compressor import (
        SIZE_OF_SDRAM_ADDRESS_IN_BYTES)


class PreAllocateForBitFieldRouterCompressor(object):
    """ Preallocates resources required for bitfield router compression.
    """

    def __call__(self, machine, sdram_to_pre_alloc_for_bit_fields,
                 pre_allocated_resources):
        """
        :param ~spinn_machine.Machine machine:
            the SpiNNaker machine as discovered
        :param int sdram_to_pre_alloc_for_bit_fields:
            SDRAM end user managed to help with bitfield compressions.
            Basically ensuring some SDRAM is available in the case where there
            is no SDRAM available to steal/use.
        :param pre_allocated_resources: other preallocated resources
        :type pre_allocated_resources:
            ~pacman.model.resources.PreAllocatedResourceContainer
        :return: preallocated resources
        :rtype: ~pacman.model.resources.PreAllocatedResourceContainer
        """

        progress_bar = ProgressBar(
            machine.n_chips,
            "Preallocating resources for bit field compressor")

        # for every Ethernet connected chip, get the resources needed by the
        # live packet gatherers
        sdrams = list()

        for chip in progress_bar.over(machine.chips):
            sdrams.append(SpecificChipSDRAMResource(
                chip,
                (SIZE_OF_SDRAM_ADDRESS_IN_BYTES * chip.n_user_processors) +
                sdram_to_pre_alloc_for_bit_fields))

        # note what has been preallocated
        allocated = PreAllocatedResourceContainer(
            specific_sdram_usage=sdrams)
        allocated.extend(pre_allocated_resources)
        return allocated



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.overrides import overrides
from spinn_front_end_common.abstract_models import (
    AbstractProvidesKeyToAtomMapping)


class ProvidesKeyToAtomMappingImpl(AbstractProvidesKeyToAtomMapping):
    __slots__ = ()

    @overrides(
        AbstractProvidesKeyToAtomMapping.routing_key_partition_atom_mapping)
    def routing_key_partition_atom_mapping(self, routing_info, partition):
        vertex_slice = partition.pre_vertex.vertex_slice
        keys = routing_info.get_keys(vertex_slice.n_atoms)
        start = vertex_slice.lo_atom
        return enumerate(keys, start)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from pacman.utilities.utility_calls import md5
from spinn_front_end_common.utilities.constants import SDP_PORTS


def get_simulation_header_array(
        binary_file_name, machine_time_step, time_scale_factor):
    """ Get data to be written to the simulation header

    :param str binary_file_name: The name of the binary of the application
    :param int machine_time_step: The time step of the simulation
    :param int time_scale_factor: The time scaling of the simulation
    :return: An array of values to be written as the simulation header
    :rtype: list(int)
    """
    # Get first 32-bits of the md5 hash of the application name
    application_name_hash = md5(os.path.splitext(binary_file_name)[0])[:8]

    # Write this to the system region (to be picked up by the simulation):
    data = list()
    data.append(int(application_name_hash, 16))
    data.append(machine_time_step * time_scale_factor)

    # add SDP port number for receiving synchronisations and new run times
    data.append(SDP_PORTS.RUNNING_COMMAND_SDP_PORT.value)

    return data



# Copyright (c) 2020 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import os
import sqlite3
import numpy
from spinn_front_end_common.utilities.sqlite_db import SQLiteDB
# import matplotlib.pyplot as plot
# import seaborn

# The types of router provenance that we'll plot
ROUTER_PLOTTABLES = (
    "Default_Routed_External_Multicast_Packets",
    "Dropped_FR_Packets",
    "Dropped_Multicast_Packets",
    "Dropped_Multicast_Packets_via_Local_Transmission",
    "Dropped_NN_Packets",
    "Dropped_P2P_Packets",
    "Dumped_from_a_Link",
    "Dumped_from_a_Processor",
    "Error",
    "External_FR_Packets",
    "External_Multicast_Packets",
    "External_NN_Packets",
    "External_P2P_Packets",
    "Local_Multicast_Packets",
    "Local_P2P_Packets",
    "Local_NN_Packets",
    "Local_FR_Packets",
    "Missed_for_Reinjection",
    "Received_for_Reinjection",
    "Reinjected",
    "Reinjection_Overflows")
SINGLE_PLOTNAME = "Plot.png"


class Plotter(object):
    __slots__ = ("cmap", "_db", "__have_insertion_order", "__verbose")

    __pyplot = None
    __seaborn = None

    def __init__(self, db_filename, verbose=False):
        self._db = SQLiteDB(db_filename, read_only=True, text_factory=str)
        self.__have_insertion_order = True
        self.__verbose = verbose
        self.cmap = "plasma"

    def __enter__(self):
        return self._db.__enter__()

    def __exit__(self, *args):
        return self._db.__exit__(*args)

    def __do_chip_query(self, description):
        # Does the query in one of two ways, depending on schema version
        with self._db.transaction() as cur:
            if self.__have_insertion_order:
                try:
                    return cur.execute("""
                        SELECT source_name AS "source", x, y,
                            description_name AS "description",
                            the_value AS "value"
                        FROM provenance_view
                        WHERE description LIKE ?
                        GROUP BY x, y, p
                        HAVING insertion_order = MAX(insertion_order)
                        """, (description, ))
                except sqlite3.OperationalError as e:
                    if "no such column: insertion_order" != str(e):
                        raise
                    self.__have_insertion_order = False
            return cur.execute("""
                SELECT source_name AS "source", x, y,
                    description_name AS "description",
                    MAX(the_value) AS "value"
                FROM provenance_view
                WHERE description LIKE ?
                GROUP BY x, y, p
                """, (description, ))

    def get_per_chip_prov_types(self):
        query = """
            SELECT DISTINCT description_name AS "description"
            FROM provenance_view
            WHERE x IS NOT NULL AND p IS NULL AND "description" IS NOT NULL
            """
        with self._db.transaction() as cur:
            return frozenset(row["description"] for row in cur.execute(query))

    def get_per_chip_prov_details(self, info):
        data = []
        xs = []
        ys = []
        src = None
        name = None
        for row in self.__do_chip_query("%" + info + "%"):
            if src is None:
                src = row["source"]
            if name is None:
                name = row["description"]
            data.append((row["x"], row["y"], row["value"]))
            xs.append(row["x"])
            ys.append(row["y"])
        ary = numpy.full((max(ys) + 1, max(xs) + 1), float("NaN"))
        for (x, y, value) in data:
            ary[y, x] = value
        return ((src + "/" + name).replace("_", " "),
                max(xs) + 1, max(ys) + 1, ary)

    def __do_sum_query(self, description):
        # Does the query in one of two ways, depending on schema version
        with self._db.transaction() as cur:
            if self.__have_insertion_order:
                try:
                    return cur.execute("""
                        SELECT "source", x, y, "description",
                            SUM("value") AS "value"
                        FROM (
                            SELECT source_name AS "source", x, y, p,
                                description_name AS "description",
                                the_value AS "value"
                            FROM provenance_view
                            WHERE description LIKE ? AND p IS NOT NULL
                            GROUP BY x, y, p
                            HAVING insertion_order = MAX(insertion_order))
                        GROUP BY x, y
                        """, (description, ))
                except sqlite3.OperationalError as e:
                    if "no such column: insertion_order" != str(e):
                        raise
                    self.__have_insertion_order = False
            return cur.execute("""
                SELECT "source", x, y, "description",
                    SUM("value") AS "value"
                FROM (
                    SELECT source_name AS "source", x, y,
                        description_name AS "description",
                        MAX(the_value) AS "value"
                    FROM provenance_view
                    WHERE description LIKE ? AND p IS NOT NULL
                    GROUP BY x, y, p)
                GROUP BY x, y
                """, (description, ))

    def get_per_core_prov_types(self):
        query = """
            SELECT DISTINCT description_name AS "description"
            FROM provenance_view
            WHERE x IS NOT NULL AND p IS NOT NULL
                AND "description" IS NOT NULL
            """
        with self._db.transaction() as cur:
            return frozenset(row["description"] for row in cur.execute(query))

    def get_sum_chip_prov_details(self, info):
        data = []
        xs = []
        ys = []
        src = None
        name = None
        for row in self.__do_sum_query("%" + info + "%"):
            if src is None:
                src = row["source"]
            if name is None:
                name = row["description"]
            data.append((row["x"], row["y"], row["value"]))
            xs.append(row["x"])
            ys.append(row["y"])
        ary = numpy.full((max(ys) + 1, max(xs) + 1), float("NaN"))
        for (x, y, value) in data:
            ary[y, x] = value
        return name.replace("_", " "), max(xs) + 1, max(ys) + 1, ary

    @classmethod
    def __plotter_apis(cls):
        # Import here because otherwise CI fails
        # pylint: disable=import-error
        if not cls.__pyplot:
            import matplotlib.pyplot as plot
            cls.__pyplot = plot
        if not cls.__seaborn:
            import seaborn
            cls.__seaborn = seaborn
        return cls.__pyplot, cls.__seaborn

    def plot_per_core_data(self, key, output_filename):
        plot, seaborn = self.__plotter_apis()
        if self.__verbose:
            print("creating " + output_filename)
        (title, width, height, data) = self.get_sum_chip_prov_details(key)
        _fig, ax = plot.subplots(figsize=(width, height))
        plot.title(title)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.axis("off")
        labels = data.astype(int)
        seaborn.heatmap(
            data, annot=labels, fmt="", square=True,
            cmap=self.cmap).invert_yaxis()
        plot.savefig(output_filename, bbox_inches='tight')
        plot.close()

    def plot_per_chip_data(self, key, output_filename):
        plot, seaborn = self.__plotter_apis()
        if self.__verbose:
            print("creating " + output_filename)
        (title, width, height, data) = self.get_per_chip_prov_details(key)
        _fig, ax = plot.subplots(figsize=(width, height))
        plot.title(title)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.axis("off")
        labels = data.astype(int)
        seaborn.heatmap(
            data, annot=labels, fmt="", square=True,
            cmap=self.cmap).invert_yaxis()
        plot.savefig(output_filename, bbox_inches='tight')
        plot.close()


def main():
    ap = argparse.ArgumentParser(
        description="Generate heat maps from SpiNNaker provenance databases.")
    ap.add_argument("-c", "--colourmap", nargs="?", default="plasma",
                    help="colour map rule for plot; default 'plasma'")
    ap.add_argument("-l", "--list", action="store_true", default=False,
                    help="list the types of metadata available")
    ap.add_argument("-q", "--quiet", action="store_true", default=False,
                    help="don't print progress information")
    ap.add_argument("-s", "--sumcores", action="store_true", default=False,
                    help="compute information by summing data from the cores "
                    "of each chip; needs a metadata_name as well unless the "
                    "--list option is also given")
    ap.add_argument("dbfile", metavar="database_file",
                    help="the provenance database to extract data from; "
                    "usually called 'provenance.sqlite3'")
    ap.add_argument("term", metavar="metadata_name", nargs="?", default=None,
                    help="the name of the metadata to plot, or a unique "
                    "fragment of it; if omitted, maps will be produced for "
                    "all the router provenance categories")
    args = ap.parse_args()

    plotter = Plotter(args.dbfile, not args.quiet)
    plotter.cmap = args.colourmap
    with plotter:
        if args.list:
            if args.sumcores:
                for term in plotter.get_per_core_prov_types():
                    print(term)
            else:
                for term in plotter.get_per_chip_prov_types():
                    print(term)
        elif args.term:
            if args.sumcores:
                plotter.plot_per_core_data(
                    args.term, os.path.abspath(SINGLE_PLOTNAME))
            else:
                plotter.plot_per_chip_data(
                    args.term, os.path.abspath(SINGLE_PLOTNAME))
        else:
            if args.sumcores:
                raise Exception(
                    "cannot use --sumcores with default router provenance")
            for term in ROUTER_PLOTTABLES:
                plotter.plot_per_chip_data(
                    term, os.path.abspath(term + ".png"))


if __name__ == "__main__":
    main()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from pacman.model.resources import (
    CoreResource, PreAllocatedResourceContainer, SpecificChipSDRAMResource)
from spinn_front_end_common.utility_models import (
    ChipPowerMonitorMachineVertex)


class PreAllocateResourcesForChipPowerMonitor(object):
    """ Adds chip power monitor resources as required for a machine.
    """

    def __call__(
            self, machine, n_samples_per_recording,
            sampling_frequency, time_scale_factor, machine_time_step,
            pre_allocated_resources):
        """
        :param ~spinn_machine.Machine machine:
            the SpiNNaker machine as discovered
        :param int n_samples_per_recording:
            how many samples between record entries
        :param int sampling_frequency: the frequency of sampling
        :param int time_scale_factor: the time scale factor
        :param int machine_time_step: the machine time step
        :param pre_allocated_resources: other preallocated resources
        :type pre_allocated_resources:
            ~pacman.model.resources.PreAllocatedResourceContainer
        :return: preallocated resources
        :rtype: ~pacman.model.resources.PreAllocatedResourceContainer
        """
        # pylint: disable=too-many-arguments

        progress_bar = ProgressBar(
            machine.n_chips, "Preallocating resources for chip power monitor")

        # store how much SDRAM the power monitor uses per core
        resources = ChipPowerMonitorMachineVertex.get_resources(
            n_samples_per_recording=n_samples_per_recording,
            sampling_frequency=sampling_frequency,
            time_scale_factor=time_scale_factor,
            time_step=machine_time_step)

        # for every Ethernet connected chip, get the resources needed by the
        # live packet gatherers
        sdrams = list()
        cores = list()
        for chip in progress_bar.over(machine.chips):
            sdrams.append(
                SpecificChipSDRAMResource(chip, resources.sdram))
            cores.append(CoreResource(chip, 1))

        # note what has been preallocated
        allocated = PreAllocatedResourceContainer(
            specific_sdram_usage=sdrams, core_resources=cores)
        allocated.extend(pre_allocated_resources)
        return allocated



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from pacman.model.constraints.key_allocator_constraints import (
    AbstractKeyAllocatorConstraint)
from spinn_front_end_common.abstract_models import (
    AbstractProvidesOutgoingPartitionConstraints,
    AbstractProvidesIncomingPartitionConstraints)


class ProcessPartitionConstraints(object):
    """ Adds constraints to partitions if the vertices at either end of the\
        partition request it.
    """

    def __call__(self, machine_graph):
        """
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
        """
        # generate progress bar
        progress = ProgressBar(
            machine_graph.n_vertices,
            "Getting constraints for machine graph")

        # iterate over each partition in the graph
        for vertex in progress.over(machine_graph.vertices):
            for partition in machine_graph.\
                    get_multicast_edge_partitions_starting_at_vertex(
                        vertex):
                self._process_partition(partition)

    @staticmethod
    def _process_partition(partition):
        """
        Process the partition by checking the pre_vertex and post vertices

        Note: The machine level is checked first and only only if that does
            not provide the api the app vertex is check next
        :param ~.OutgoingEdgePartition partition:
        """
        vertex = partition.pre_vertex
        # add_vertex_constraints
        for constraint in vertex.constraints:
            if isinstance(constraint, AbstractKeyAllocatorConstraint):
                partition.add_constraint(constraint)

        # call get_outgoing_partition_constraints method on pre_vertex
        if isinstance(vertex, AbstractProvidesOutgoingPartitionConstraints):
            partition.add_constraints(
                vertex.get_outgoing_partition_constraints(partition))
        else:
            vertex = vertex.app_vertex
            if isinstance(vertex,
                          AbstractProvidesOutgoingPartitionConstraints):
                partition.add_constraints(
                    vertex.get_outgoing_partition_constraints(partition))

        # call get_incoming_partition_constraints on post_vertex
        for edge in partition.edges:
            post_vertex = edge.post_vertex
            if isinstance(post_vertex,
                          AbstractProvidesIncomingPartitionConstraints):
                partition.add_constraints(
                    post_vertex.get_incoming_partition_constraints(partition))
            elif edge.app_edge:
                post_vertex = edge.app_edge.post_vertex
                if isinstance(post_vertex,
                              AbstractProvidesIncomingPartitionConstraints):
                    partition.add_constraints(
                        post_vertex.get_incoming_partition_constraints(
                            partition))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sqlite3
import time
from spinn_utilities.abstract_context_manager import AbstractContextManager
from spinn_utilities.overrides import overrides
from spinn_front_end_common.utilities.sqlite_db import SQLiteDB
from .abstract_database import AbstractDatabase

_DDL_FILE = os.path.join(os.path.dirname(__file__), "db.sql")
_SECONDS_TO_MICRO_SECONDS_CONVERSION = 1000


def _timestamp():
    return int(time.time() * _SECONDS_TO_MICRO_SECONDS_CONVERSION)


class SqlLiteDatabase(SQLiteDB, AbstractContextManager):
    """ Specific implementation of the Database for SQLite 3.

    .. note::
        *Not thread safe on the same database file!*
        Threads can access different DBs just fine.
    """

    __slots__ = []

    def __init__(self, database_file=None):
        """
        :param str database_file:
            The name of a file that contains (or will contain) an SQLite
            database holding the data. If omitted, an unshared in-memory
            database will be used.
        """
        super().__init__(database_file, ddl_file=_DDL_FILE)

    @overrides(AbstractDatabase.clear)
    def clear(self):
        with self.transaction() as cursor:
            cursor.execute(
                """
                UPDATE region SET
                    content = CAST('' AS BLOB), content_len = 0,
                    fetches = 0, append_time = NULL
                """)
            cursor.execute("DELETE FROM region_extra")

    @overrides(AbstractDatabase.clear_region)
    def clear_region(self, x, y, p, region):
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT region_id FROM region_view
                    WHERE x = ? AND y = ? AND processor = ?
                        AND local_region_index = ? AND fetches > 0 LIMIT 1
                    """, (x, y, p, region)):
                locus = (row["region_id"], )
                break
            else:
                return False
            cursor.execute(
                """
                UPDATE region SET
                    content = CAST('' AS BLOB), content_len = 0,
                    fetches = 0, append_time = NULL
                WHERE region_id = ?
                """, locus)
            cursor.execute(
                """
                DELETE FROM region_extra WHERE region_id = ?
                """, locus)
            return True

    def __read_contents(self, cursor, x, y, p, region):
        """
        :param ~sqlite3.Cursor cursor:
        :param int x:
        :param int y:
        :param int p:
        :param int region:
        :rtype: memoryview
        """
        for row in cursor.execute(
                """
                SELECT region_id, content, have_extra
                FROM region_view
                WHERE x = ? AND y = ? AND processor = ?
                    AND local_region_index = ? LIMIT 1
                """, (x, y, p, region)):
            r_id, data, extra = (
                row["region_id"], row["content"], row["have_extra"])
            break
        else:
            raise LookupError("no record for region ({},{},{}:{})".format(
                x, y, p, region))
        if extra:
            c_buffer = None
            for row in cursor.execute(
                    """
                    SELECT r.content_len + (
                        SELECT SUM(x.content_len)
                        FROM region_extra AS x
                        WHERE x.region_id = r.region_id) AS len
                    FROM region AS r WHERE region_id = ? LIMIT 1
                    """, (r_id, )):
                c_buffer = bytearray(row["len"])
                c_buffer[:len(data)] = data
            idx = len(data)
            for row in cursor.execute(
                    """
                    SELECT content FROM region_extra
                    WHERE region_id = ? ORDER BY extra_id ASC
                    """, (r_id, )):
                item = row["content"]
                c_buffer[idx:idx + len(item)] = item
                idx += len(item)
            data = c_buffer
        return memoryview(data)

    @staticmethod
    def __get_core_id(cursor, x, y, p):
        """
        :param ~sqlite3.Cursor cursor:
        :param int x:
        :param int y:
        :param int p:
        :rtype: int
        """
        for row in cursor.execute(
                """
                SELECT core_id FROM region_view
                WHERE x = ? AND y = ? AND processor = ?
                LIMIT 1
                """, (x, y, p)):
            return row["core_id"]
        cursor.execute(
            """
            INSERT INTO core(x, y, processor) VALUES(?, ?, ?)
            """, (x, y, p))
        return cursor.lastrowid

    def __get_region_id(self, cursor, x, y, p, region):
        """
        :param ~sqlite3.Cursor cursor:
        :param int x:
        :param int y:
        :param int p:
        :param int region:
        """
        for row in cursor.execute(
                """
                SELECT region_id FROM region_view
                WHERE x = ? AND y = ? AND processor = ?
                    AND local_region_index = ?
                LIMIT 1
                """, (x, y, p, region)):
            return row["region_id"]
        core_id = self.__get_core_id(cursor, x, y, p)
        cursor.execute(
            """
            INSERT INTO region(
                core_id, local_region_index, content, content_len, fetches)
            VALUES(?, ?, CAST('' AS BLOB), 0, 0)
            """, (core_id, region))
        return cursor.lastrowid

    @overrides(AbstractDatabase.store_data_in_region_buffer)
    def store_data_in_region_buffer(self, x, y, p, region, missing, data):
        # pylint: disable=too-many-arguments
        # TODO: Use missing
        datablob = sqlite3.Binary(data)
        with self.transaction() as cursor:
            region_id = self.__get_region_id(cursor, x, y, p, region)
            if self.__use_main_table(cursor, region_id):
                cursor.execute(
                    """
                    UPDATE region SET
                        content = CAST(? AS BLOB),
                        content_len = ?,
                        fetches = fetches + 1,
                        append_time = ?
                    WHERE region_id = ?
                    """, (datablob, len(data), _timestamp(), region_id))
            else:
                cursor.execute(
                    """
                    UPDATE region SET
                        fetches = fetches + 1,
                        append_time = ?
                    WHERE region_id = ?
                    """, (_timestamp(), region_id))
                assert cursor.rowcount == 1
                cursor.execute(
                    """
                    INSERT INTO region_extra(
                        region_id, content, content_len)
                    VALUES (?, CAST(? AS BLOB), ?)
                    """, (region_id, datablob, len(data)))
            assert cursor.rowcount == 1

    def __use_main_table(self, cursor, region_id):
        """
        :param ~sqlite3.Cursor cursor:
        :param int region_id:
        """
        for row in cursor.execute(
                """
                SELECT COUNT(*) AS existing FROM region
                WHERE region_id = ? AND fetches = 0
                LIMIT 1
                """, (region_id, )):
            existing = row["existing"]
            return existing == 1
        return False

    @overrides(AbstractDatabase.get_region_data)
    def get_region_data(self, x, y, p, region):
        try:
            with self.transaction() as cursor:
                data = self.__read_contents(cursor, x, y, p, region)
                # TODO missing data
                return data, False
        except LookupError:
            return memoryview(b''), True



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


class AbstractChangableAfterRun(object, metaclass=AbstractBase):
    """ An item that can be changed after a call to run, the changes to which\
        might or might not require mapping or data generation.
    """

    __slots__ = ()

    @property
    def requires_mapping(self):
        """ True if changes that have been made require that mapping be\
            performed.  By default this returns False but can be overridden to\
            indicate changes that require mapping.

        :rtype: bool
        """
        return False

    @property
    def requires_data_generation(self):
        """ True if changes that have been made require that data generation\
            be performed.  By default this returns False but can be overridden\
            to indicate changes that require data regeneration.

        :rtype: bool
        """
        return False

    @abstractmethod
    def mark_no_changes(self):
        """ Marks the point after which changes are reported, so that new\
            changes can be detected before the next check.
        """
